"""
Prototype functions for station location and timing summary
Provides an example of how to access the information from a Station object
This module will contain methods to summarize a single Station
"""
from typing import List, Optional
import math

from dataclasses import dataclass
from dataclasses_json import dataclass_json
import numpy as np

from redvox.common.station import Station
from redvox.api1000.wrapped_redvox_packet.sensors.location import LocationProvider


LOCATION_PROVIDER_IDX: int = 10
LATITUDE_IDX: int = 1
LONGITUDE_IDX: int = 2
ALTITUDE_IDX: int = 3


@dataclass_json
@dataclass
class LocationSummary:
    """
    Summary of a Location.  Includes:

    * Location sensor's name
    * Source of data (provider)
    * Number of points
    * The mean and standard deviation of latitude, longitude, and altitude.

    Properties:
        name: str, name of the Location sensor

        provider: str, name of the provider for location data

        latitude_mean: float, mean of the latitude values.  Default 0.0

        latitude_standard_deviation: float, standard deviation of the latitude values.  Default 0.0

        longitude_mean: float, mean of the longitude values.  Default 0.0

        longitude_standard_deviation: float, standard deviation of the longitude values.  Default 0.0

        altitude_mean: float, mean of the altitude values.  Default 0.0

        altitude_standard_deviation: float, standard deviation of the altitude values.  Default 0.0

        num_pts: int, the number of data points in the summary.  Default 0
    """

    name: str
    provider: str
    latitude_mean: float = 0.0
    latitude_standard_deviation: float = 0.0
    longitude_mean: float = 0.0
    longitude_standard_deviation: float = 0.0
    altitude_mean: float = 0.0
    altitude_standard_deviation: float = 0.0
    num_pts: int = 0

    @staticmethod
    def from_station(stn: Station) -> Optional[List["LocationSummary"]]:
        """
        :param stn: Station to get data from
        :return: List of LocationSummary grouped by provider or None
        """
        loc = stn.find_loc_for_stats()
        if loc:
            samples: np.ndarray = loc.samples()

            if loc.num_samples() == 1:
                return [
                    LocationSummary(
                        loc.name,
                        LocationProvider(int(samples[LOCATION_PROVIDER_IDX][0])).name,
                        samples[LATITUDE_IDX][0],
                        0.0,
                        samples[LONGITUDE_IDX][0],
                        0.0,
                        samples[ALTITUDE_IDX][0],
                        0.0,
                        1,
                    )
                ]

            # for each provider, create a new entry in the dictionary or append new data
            loc_prov_to_data: dict = {}
            lat_samples: np.ndarray = samples[LATITUDE_IDX]
            lng_samples: np.ndarray = samples[LONGITUDE_IDX]
            for j in range(min(len(lat_samples), len(lng_samples))):
                if (not math.isnan(lat_samples[j])) and (not math.isnan(lng_samples[j])):
                    prov = LocationProvider(int(samples[LOCATION_PROVIDER_IDX][j])).name
                    data = np.array(
                        [
                            [samples[LATITUDE_IDX][j]],
                            [samples[LONGITUDE_IDX][j]],
                            [samples[ALTITUDE_IDX][j]],
                        ]
                    )
                    loc_prov_to_data[prov] = (
                        data
                        if prov not in loc_prov_to_data.keys()
                        else np.concatenate([loc_prov_to_data[prov], data], 1)
                    )

            loc_sums = []
            for provider, loc_data in loc_prov_to_data.items():
                loc_sums.append(
                    LocationSummary(
                        loc.name,
                        provider,
                        loc_data[0].mean(),
                        loc_data[0].std(),
                        loc_data[1].mean(),
                        loc_data[1].std(),
                        loc_data[2].mean(),
                        loc_data[2].std(),
                        len(loc_data[0]),
                    )
                )
            return loc_sums
        return None



"""
This module provides tools to fill gaps and pad the ends of timestamp arrays
"""
import sys
from typing import List, Tuple, Optional, Dict
import enum
from math import modf
from dataclasses import dataclass, field

from dataclasses_json import dataclass_json
import numpy as np
import pyarrow as pa
import pyarrow.compute as pc

from redvox.common import date_time_utils as dtu
from redvox.common.errors import RedVoxExceptions
from redvox.api1000.wrapped_redvox_packet.sensors.audio import AudioCodec
from redvox.api1000.wrapped_redvox_packet.sensors.location import LocationProvider
from redvox.api1000.wrapped_redvox_packet.sensors.image import ImageCodec
from redvox.api1000.wrapped_redvox_packet.station_information import (
    NetworkType,
    PowerState,
    CellServiceState,
    WifiWakeLock,
    ScreenState,
)

# percent of packet duration/sample rate required for gap to be considered a whole unit
DEFAULT_GAP_UPPER_LIMIT: float = 0.8
# percent of packet duration/sample rate required for gap to be considered nothing
DEFAULT_GAP_LOWER_LIMIT: float = 0.02
# columns for audio table
AUDIO_DF_COLUMNS = ["timestamps", "unaltered_timestamps", "microphone"]
# columns that cannot be interpolated
NON_INTERPOLATED_COLUMNS = ["compressed_audio", "image"]
# columns that are not numeric but can be interpolated
NON_NUMERIC_COLUMNS = [
    "location_provider",
    "image_codec",
    "audio_codec",
    "network_type",
    "power_state",
    "cell_service",
    "wifi_wake_lock",
    "screen_state",
]


# noinspection Mypy,DuplicatedCode
class DataPointCreationMode(enum.Enum):
    """
    Type of data point to create
    """

    NAN: int = 0
    COPY: int = 1
    INTERPOLATE: int = 2

    @staticmethod
    def list_names() -> List[str]:
        return [n.name for n in DataPointCreationMode]


@dataclass_json
@dataclass
class AudioWithGaps:
    """
    Represents methods of reconstructing audio data with or without gaps in it

    Properties:
        sample_interval_micros: microseconds between sample points

        metadata: list of start times in microseconds since epoch UTC and the data to add

        gaps: the list of start and end points of gaps (the start and end are actual data points)

        errors: the errors encountered while getting the data
    """

    sample_interval_micros: float
    metadata: Optional[List[Tuple[float, pa.Table]]] = None
    gaps: List[Tuple[float, float]] = field(default_factory=lambda: [])
    errors: RedVoxExceptions = field(default_factory=lambda: RedVoxExceptions("AudioWithGaps"))

    def create_timestamps(self) -> pa.Table:
        """
        :return: converts the audio metadata into a data table
        """
        result_array = [[], [], []]
        for m in self.metadata:
            timestamps = calc_evenly_sampled_timestamps(m[0], m[1].num_rows, self.sample_interval_micros)
            result_array[0].extend(timestamps)
            result_array[1].extend(timestamps)
            result_array[2].extend(m[1]["microphone"].to_numpy())
        for gs, ge in self.gaps:
            fractional, whole = modf((ge - gs) / self.sample_interval_micros)
            num_samples = int((whole - 1) if fractional < DEFAULT_GAP_LOWER_LIMIT else whole)
            timestamps = calc_evenly_sampled_timestamps(
                gs + self.sample_interval_micros, num_samples, self.sample_interval_micros
            )
            gap_array = [timestamps, np.full(len(timestamps), np.nan)]
            result_array[0].extend(gap_array[0])
            result_array[1].extend(gap_array[0])
            result_array[2].extend(gap_array[1])
        ptable = pa.Table.from_pydict(dict(zip(AUDIO_DF_COLUMNS, result_array)))
        return pc.take(ptable, pc.sort_indices(ptable, sort_keys=[("timestamps", "ascending")]))

    def add_error(self, error: str):
        """
        add an error to the result
        :param error: error message to add
        """
        self.errors.append(error)


def calc_evenly_sampled_timestamps(start: float, samples: int, sample_interval_micros: float) -> np.array:
    """
    given a start time, calculates samples amount of evenly spaced timestamps at rate_hz

    :param start: float, start timestamp in microseconds
    :param samples: int, number of samples
    :param sample_interval_micros: float, sample interval in microseconds
    :return: np.array with number of samples timestamps, evenly spaced starting at start
    """
    return start + (np.arange(0, samples) * sample_interval_micros)


def check_gap_list(
    gaps: List[Tuple[float, float]], start_timestamp: float = None, end_timestamp: float = None
) -> List[Tuple[float, float]]:
    """
    removes any gaps where end time <= start time, consolidates overlapping gaps, and ensures that no gap
    starts or ends before start_timestamp and starts or ends after end_timestamp.  All timestamps are in
    microseconds since epoch UTC

    :param gaps: list of gaps to check
    :param start_timestamp: lowest possible timestamp for a gap to start at
    :param end_timestamp: lowest possible timestamp for a gap to end at
    :return: list of correct, valid gaps
    """
    return_gaps: List[Tuple[float, float]] = []
    for gap in gaps:
        if start_timestamp:
            gap = (np.max([start_timestamp, gap[0]]), np.max([start_timestamp, gap[1]]))
        if end_timestamp:
            gap = (np.min([end_timestamp, gap[0]]), np.min([end_timestamp, gap[1]]))
        if gap[0] < gap[1]:
            if len(return_gaps) < 1:
                return_gaps.append(gap)
            for a, r_g in enumerate(return_gaps):
                if (gap[0] < r_g[0] and gap[1] < r_g[0]) or (gap[0] > r_g[1] and gap[1] > r_g[1]):
                    return_gaps.append(gap)
                    break
                else:
                    if gap[0] < r_g[0] < gap[1]:
                        r_g = (gap[0], r_g[1])
                    if gap[0] < r_g[1] < gap[1]:
                        r_g = (r_g[0], gap[1])
                    return_gaps[a] = r_g
    return return_gaps


def fill_gaps(
    arrow_df: pa.Table, gaps: List[Tuple[float, float]], sample_interval_micros: float, fill_mode: str = "nan"
) -> Tuple[pa.Table, List[Tuple[float, float]]]:
    """
    fills gaps in the table with np.nan or interpolated values by interpolating timestamps based on the
    calculated sample interval

    :param arrow_df: pyarrow table with data.  first column is "timestamps"
    :param gaps: list of tuples of known non-inclusive start and end timestamps of the gaps
    :param sample_interval_micros: known sample interval of the data points
    :param fill_mode: must be one of: "nan", "interpolate", or "copy".  Other inputs result in "nan".  Default "nan".
                        Will convert input to lowercase.
    :return: table without gaps and the list of gaps
    """
    # extract the necessary information to compute gap size and gap timestamps
    data_time_stamps = arrow_df["timestamps"].to_numpy()
    if len(data_time_stamps) > 1:
        data_duration = data_time_stamps[-1] - data_time_stamps[0]
        expected_samples = (
            np.floor(data_duration / sample_interval_micros)
            + (1 if data_duration % sample_interval_micros >= sample_interval_micros * DEFAULT_GAP_UPPER_LIMIT else 0)
        ) + 1
        if expected_samples > len(data_time_stamps):
            if fill_mode.lower() == "copy":
                pcm = DataPointCreationMode["COPY"]
            elif fill_mode.lower() == "interpolate":
                pcm = DataPointCreationMode["INTERPOLATE"]
            else:
                pcm = DataPointCreationMode["NAN"]
            # make it safe to alter the gap values
            my_gaps = check_gap_list(gaps, data_time_stamps[0], data_time_stamps[-1])
            for gap in my_gaps:
                # if timestamps are around gaps, we have to update the values
                before_start = np.argwhere([t <= gap[0] for t in data_time_stamps])
                after_end = np.argwhere([t >= gap[1] for t in data_time_stamps])
                if len(before_start) > 0:
                    before_start = before_start[-1][0]
                    # sim = gap[0] - data_time_stamps[before_start]
                    # result_df = add_data_points_to_df(result_df, before_start, sim, point_creation_mode=pcm)
                    gap = (data_time_stamps[before_start], gap[1])
                else:
                    before_start = None
                if len(after_end) > 0:
                    after_end = after_end[0][0]
                    # sim = gap[1] - data_time_stamps[after_end]
                    gap = (gap[0], data_time_stamps[after_end])
                else:
                    after_end = None
                num_new_points = int((gap[1] - gap[0]) / sample_interval_micros) - 1
                if before_start is not None:
                    arrow_df = add_data_points_to_df(
                        arrow_df, before_start, sample_interval_micros, num_new_points, pcm
                    )
                elif after_end is not None:
                    arrow_df = add_data_points_to_df(arrow_df, after_end, -sample_interval_micros, num_new_points, pcm)
        indic = pc.sort_indices(arrow_df, sort_keys=[("timestamps", "ascending")])
        return arrow_df.take(indic), gaps
    return arrow_df, gaps


def fill_audio_gaps(
    packet_data: List[Tuple[float, pa.Table]],
    sample_interval_micros: float,
    gap_lower_limit: float = DEFAULT_GAP_LOWER_LIMIT,
) -> AudioWithGaps:
    """
    fills gaps in the table with np.nan by interpolating timestamps based on the expected sample interval
      * ignores gaps with duration less than or equal to packet length * gap_lower_limit

    :param packet_data: list of tuples, each tuple containing two pieces of packet information:
                        packet_start_timestamps; float of packet start timestamp in microseconds
                        and audio_data; pa.Table of data points
    :param sample_interval_micros: sample interval in microseconds
    :param gap_lower_limit: percentage of packet length required to disregard gap, default DEFAULT_GAP_LOWER_LIMIT
    :return: AudioWithGaps object that contains a list of timestamps of the non-inclusive start and end of the gaps
                and other information to recreate the audio record
    """
    last_data_timestamp = packet_data[0][0]
    gaps = []
    result = AudioWithGaps(sample_interval_micros, packet_data)
    for packet in packet_data:
        samples_in_packet = packet[1].num_rows
        start_ts = packet[0]
        packet_length = sample_interval_micros * samples_in_packet
        # check if start_ts is close to the last timestamp in data_timestamps
        last_timestamp_diff = start_ts - last_data_timestamp
        if last_timestamp_diff > gap_lower_limit * packet_length:
            gap_start = last_data_timestamp - sample_interval_micros
            last_data_timestamp = start_ts
            gaps.append((gap_start, last_data_timestamp))
        elif last_timestamp_diff < -gap_lower_limit * packet_length:
            result.add_error(
                f"Packet start timestamp: {dtu.microseconds_to_seconds(start_ts)} "
                f"is before last timestamp of previous "
                f"packet: {dtu.microseconds_to_seconds(last_data_timestamp - sample_interval_micros)}"
            )
        last_data_timestamp += samples_in_packet * sample_interval_micros
    result.gaps = gaps
    return result


def add_data_points_to_df(
    data_table: pa.Table,
    start_index: int,
    sample_interval_micros: float,
    num_samples_to_add: int = 1,
    point_creation_mode: DataPointCreationMode = DataPointCreationMode.COPY,
) -> pa.Table:
    """
    adds data points to the end of the table, starting from the index specified.
        Note:
            * table must not be empty
            * start_index must be non-negative and less than the length of table
            * num_samples_to_add must be greater than 0
            * sample_interval_micros cannot be 0
            * points are added onto the end and the result is not sorted
        Options for point_creation_mode are:
            * NAN: default values and nans
            * COPY: copies of the start data point
            * INTERPOLATE: interpolated values between start data point and adjacent point

    :param data_table: pyarrow table to add dataless timestamps to
    :param start_index: index of the table to use as starting point for creating new values
    :param sample_interval_micros: sample interval in microseconds of the timestamps; use negative values to
                                    add points before the start_index
    :param num_samples_to_add: the number of timestamps to create, default 1
    :param point_creation_mode: the mode of point creation to use
    :return: updated table with synthetic data points
    """
    if (
        len(data_table) > start_index
        and len(data_table) > 0
        and num_samples_to_add > 0
        and sample_interval_micros != 0.0
    ):
        start_timestamp = data_table["timestamps"][start_index].as_py()
        # create timestamps for every point that needs to be added
        new_timestamps = start_timestamp + np.arange(1, num_samples_to_add + 1) * sample_interval_micros
        if point_creation_mode == DataPointCreationMode.COPY:
            # copy the start point
            copy_row = data_table.slice(start_index, 1).to_pydict()
            for t in new_timestamps:
                copy_row["timestamps"] = [t]
                # for k in copy_row.keys():
                #     new_dict[k].append(copy_row[k])
            empty_df = pa.Table.from_pydict(copy_row)
        elif point_creation_mode == DataPointCreationMode.INTERPOLATE:
            # use the start point and the next point as the edges for interpolation
            start_point = data_table.slice(start_index, 1).to_pydict()
            numeric_start = start_point[
                [col for col in data_table.schema.names if col not in NON_INTERPOLATED_COLUMNS + NON_NUMERIC_COLUMNS]
            ]
            non_numeric_start = start_point[[col for col in data_table.schema.names if col in NON_NUMERIC_COLUMNS]]
            end_point = data_table.slice(start_index + (1 if sample_interval_micros > 0 else -1), 1).to_pydict()
            numeric_end = end_point[
                [col for col in data_table.schema.names if col not in NON_INTERPOLATED_COLUMNS + NON_NUMERIC_COLUMNS]
            ]
            non_numeric_end = end_point[[col for col in data_table.schema.names if col in NON_NUMERIC_COLUMNS]]
            if np.abs(start_point["timestamps"] - new_timestamps[0]) <= np.abs(
                end_point["timestamps"] - new_timestamps[0]
            ):
                non_numeric_diff = non_numeric_start
            else:
                non_numeric_diff = non_numeric_end
            numeric_diff = numeric_end - numeric_start
            numeric_diff = (numeric_diff / numeric_diff["timestamps"]) * (
                new_timestamps - numeric_start
            ) + numeric_start
            if sys.version_info[0] > 3 or (sys.version_info[0] == 3 and sys.version_info[1] >= 9):
                # merge dicts (python 3.9):
                empty_df = pa.Table.from_pydict(numeric_diff | non_numeric_diff)
            else:
                # merge dicts (python 3.5 to 3.8)
                empty_df = pa.Table.from_pydict({**numeric_diff, **non_numeric_diff})
        else:
            # add nans and defaults
            empty_dict: Dict[str, List] = {}
            for k in data_table.schema.names:
                empty_dict[k] = []
            for column_index in data_table.schema.names:
                if column_index == "timestamps":
                    empty_dict[column_index] = new_timestamps
                elif column_index == "location_provider":
                    empty_dict[column_index] = [LocationProvider["UNKNOWN"].value for i in range(num_samples_to_add)]
                elif column_index == "image_codec":
                    empty_dict[column_index] = [ImageCodec["UNKNOWN"].value for i in range(num_samples_to_add)]
                elif column_index == "audio_codec":
                    empty_dict[column_index] = [AudioCodec["UNKNOWN"].value for i in range(num_samples_to_add)]
                elif column_index == "network_type":
                    empty_dict[column_index] = [NetworkType["UNKNOWN_NETWORK"].value for i in range(num_samples_to_add)]
                elif column_index == "power_state":
                    empty_dict[column_index] = [
                        PowerState["UNKNOWN_POWER_STATE"].value for i in range(num_samples_to_add)
                    ]
                elif column_index == "cell_service":
                    empty_dict[column_index] = [CellServiceState["UNKNOWN"].value for i in range(num_samples_to_add)]
                elif column_index == "wifi_wake_lock":
                    empty_dict[column_index] = [WifiWakeLock["NONE"].value for i in range(num_samples_to_add)]
                elif column_index == "screen_state":
                    empty_dict[column_index] = [
                        ScreenState["UNKNOWN_SCREEN_STATE"].value for i in range(num_samples_to_add)
                    ]
                else:
                    empty_dict[column_index] = np.full(num_samples_to_add, np.nan).tolist()
            empty_df = pa.Table.from_pydict(empty_dict)
        data_table = pa.concat_tables([data_table, empty_df])

    return data_table



"""
This module provides methods and datatypes for the efficient extraction and manipulation of machine learning data stored
 in RedVox packet EventStreams.
"""

from dataclasses import dataclass
from dataclasses_json import dataclass_json
from enum import Enum
from math import isfinite
from typing import Optional, Dict, List

import numpy as np
from redvox.common.errors import RedVoxError

from redvox.api1000.common.generic import ProtoRepeatedMessage
from redvox.api1000.wrapped_redvox_packet.event_streams import EventStream, Event
from redvox.api1000.wrapped_redvox_packet.wrapped_packet import WrappedRedvoxPacketM

ML_CLASS_PREFIX: str = "class_"
ML_SCORE_PREFIX: str = "score_"
ML_EVENT_STREAM_NAME: str = "inference"
ML_METADATA_MODEL_VERSION_KEY: str = "model_version"
ML_METADATA_INPUT_SAMPLES_PER_HOP_KEY: str = "input_samples_per_hop"
ML_METADATA_INPUT_SAMPLE_RATE_KEY: str = "input_sample_rate"
ML_METADATA_INPUT_SAMPLES_PER_WINDOW_KEY: str = "input_samples_per_window"


class MlError(RedVoxError):
    def __init__(self, msg: str):
        super().__init__(f"MlError: {msg}")


@dataclass_json
@dataclass
class ExtractedMl:
    """
    Contains the extracted ML classes and scores as well as metadata relating to the utilized model.
    """

    metadata: "MlMetadata"
    windows: List["MlWindow"]

    def sort(self, sort_by: "SortBy") -> "ExtractedMl":
        """
        Sorts the labels in ascending or descending order by either score of class name.
        :param sort_by: The sort operation to use.
        :return: An updated instance of ExtractedMl.
        """
        self.windows = list(map(lambda window: window.sort(sort_by), self.windows))
        return self

    def prune_zeros(self) -> "ExtractedMl":
        """
        Removes labels that have a score of 0.
        :return: An updated instance of ExtractedMl.
        """
        self.windows = list(map(lambda window: window.prune_zeros(), self.windows))
        return self

    def prune_lt(self, min_v: float) -> "ExtractedMl":
        """
        Prunes labels with score less than the provided minimum.
        :param min_v: The minimum acceptable label score.
        :return: An updated instance of ExtractedMl.
        """
        if min_v <= 0:
            raise MlError(f"min_v={min_v} must be > 0")

        self.windows = list(map(lambda window: window.prune_lt(min_v), self.windows))
        return self

    def retain_top(self, n: int) -> "ExtractedMl":
        """
        Sorts labels in descending order by score and only keeps up to the top n labels.
        :param n: The number of labels to keep.
        :return: An updated instance of ExtractedMl.
        """
        if n <= 0:
            raise MlError(f"n={n} must be > 0")

        self.windows = list(map(lambda window: window.retain_top(n), self.windows))
        return self


@dataclass_json
@dataclass
class MlMetadata:
    """
    Metadata relating to the utilized ML model.
    """

    model_name: str
    model_version: str
    input_samples_per_hop: int
    input_sample_rate: int
    input_samples_per_window: int


@dataclass_json
@dataclass
class MlWindow:
    """
    Labels from a single time window.
    """

    timestamp: int
    labels: List["Label"]

    def sort(self, sort_by: "SortBy") -> "MlWindow":
        """
        Sorts the labels in ascending or descending order by either score or class name.
        :param sort_by: The sort operation to use.
        :return: An updated instance of MlWindow.
        """
        if sort_by is SortBy.CLASS_ASC:
            self.labels = sorted(self.labels, key=lambda label: label.class_name)
        elif sort_by is SortBy.CLASS_DESC:
            self.labels = sorted(self.labels, key=lambda label: label.class_name, reverse=True)
        elif sort_by is SortBy.SCORE_ASC:
            self.labels = sorted(self.labels, key=lambda label: label.score)
        else:
            self.labels = sorted(self.labels, key=lambda label: label.score, reverse=True)

        return self

    def prune_zeros(self) -> "MlWindow":
        """
        Removes labels that have a score of 0.
        :return: An updated instance of MlWindow.
        """
        self.labels = list(filter(lambda label: label.score > 0, self.labels))
        return self

    def prune_lt(self, min_v: float) -> "MlWindow":
        """
        Prunes labels with score less than the provided minimum.
        :param min_v: The minimum acceptable label score.
        :return: An updated instance of MlWindow.
        """
        if min_v <= 0:
            raise MlError(f"min_v={min_v} must be > 0")

        self.labels = list(filter(lambda label: label.score >= min_v, self.labels))
        return self

    def retain_top(self, n: int) -> "MlWindow":
        """
        Sorts labels in descending order by score and only keeps up to the top n labels.
        :param n: The number of labels to keep.
        :return: An updated instance of MlWindow.
        """
        if n <= 0:
            raise MlError(f"n={n} must be > 0")

        sorted_window: MlWindow = self.sort(SortBy.SCORE_DESC)
        sorted_window.labels = sorted_window.labels[:n]
        return sorted_window


@dataclass_json
@dataclass
class Label:
    """
    A pair containing a class name and a score.
    """

    class_name: str
    score: float


class SortBy(Enum):
    """
    An enumeration that represents the valid label sorting methods.
    """

    SCORE_ASC: int = 1
    SCORE_DESC: int = 2
    CLASS_ASC: int = 3
    CLASS_DESC: int = 4


def extract_ml_metadata(stream: EventStream) -> MlMetadata:
    """
    Extracts ML metadata from an event stream.
    :param stream: The event stream to extract the ML metadata from.
    :return: An instance of MlMetadata.
    """
    if stream.get_name() != ML_EVENT_STREAM_NAME:
        raise MlError(f"Invalid ML event stream name={stream.get_name()} != {ML_EVENT_STREAM_NAME}")

    if stream.get_events().get_count() == 0:
        raise MlError("ML EventStream contains 0 Events")

    name: str = stream.get_events().get_values()[0].get_description()
    meta: Dict[str, str] = stream.get_metadata().get_metadata()

    for key in [
        ML_METADATA_MODEL_VERSION_KEY,
        ML_METADATA_INPUT_SAMPLES_PER_HOP_KEY,
        ML_METADATA_INPUT_SAMPLE_RATE_KEY,
        ML_METADATA_INPUT_SAMPLES_PER_WINDOW_KEY,
    ]:
        if key not in meta:
            raise MlError(f"Missing required ML metadata key={key}")

    try:
        return MlMetadata(
            name,
            meta[ML_METADATA_MODEL_VERSION_KEY],
            int(meta[ML_METADATA_INPUT_SAMPLES_PER_HOP_KEY]),
            int(meta[ML_METADATA_INPUT_SAMPLE_RATE_KEY]),
            int(meta[ML_METADATA_INPUT_SAMPLES_PER_WINDOW_KEY]),
        )
    except ValueError:
        raise MlError("Could not parse ML metadata")


def find_ml_event_stream(packet: WrappedRedvoxPacketM) -> Optional[EventStream]:
    """
    Attempts to find an event stream with ML data.
    :param packet: The packet to search in.
    :return: An instance of the matching event stream or None.
    """
    streams: ProtoRepeatedMessage = packet.get_event_streams()

    stream: EventStream
    for stream in streams.get_values():
        if stream.get_name() == ML_EVENT_STREAM_NAME:
            return stream

    return None


def label_at(str_payload: Dict[str, str], num_payload: Dict[str, float], idx: int) -> Label:
    """
    Finds the label and score in the event payloads.
    :param str_payload: The event string payload.
    :param num_payload: The event numeric payload.
    :param idx: The index of the class and score.
    :return: An instance of a Label.
    """
    class_key: str = f"{ML_CLASS_PREFIX}{idx}"
    score_key: str = f"{ML_SCORE_PREFIX}{idx}"

    if class_key not in str_payload:
        raise MlError(f"Missing required class_key={class_key}")

    if score_key not in num_payload:
        raise MlError(f"Missing required score_key={score_key}")

    class_name: str = str_payload[class_key]
    score: float = num_payload[score_key]

    if not isfinite(score):
        raise MlError(f"Invalid non-finite score={score}")

    return Label(class_name, score)


def extract_ml_windows(stream: EventStream) -> List[MlWindow]:
    """
    Extracts ML windows from an event stream.
    :param stream: The stream to extract windows from.
    :return: A list of ML windows.
    """
    if stream.get_name() != ML_EVENT_STREAM_NAME:
        raise MlError(f"Invalid ML event stream name={stream.get_name()} != {ML_EVENT_STREAM_NAME}")

    timestamps: np.ndarray = stream.get_timestamps().get_timestamps()
    windows: List[MlWindow] = []

    events: List[Event] = stream.get_events().get_values()

    if len(timestamps) != len(events):
        raise MlError(f"len(timestamps={len(timestamps)}) != len(events={len(events)})")

    idx_window: int
    timestamp: float
    for idx_window, timestamp in enumerate(timestamps):
        event: Event = events[idx_window]
        str_payload: Dict[str, str] = event.get_string_payload().get_metadata()
        num_payload: Dict[str, float] = event.get_numeric_payload().get_metadata()

        labels: List[Label] = []
        for i in range(len(str_payload)):
            label: Label = label_at(str_payload, num_payload, i)
            labels.append(label)

        window: MlWindow = MlWindow(int(round(timestamp)), labels)
        windows.append(window)

    return windows


def extract_ml_from_event_stream(stream: EventStream) -> ExtractedMl:
    """
    Extract ML parameters from an event stream.
    :param stream: The event stream to extract ML parameters from.
    :return: The extracted ML parameters or None.
    """
    if stream.get_name() != ML_EVENT_STREAM_NAME:
        raise MlError(f"Invalid ML event stream name={stream.get_name()} != {ML_EVENT_STREAM_NAME}")

    metadata: MlMetadata = extract_ml_metadata(stream)
    windows: List[MlWindow] = extract_ml_windows(stream)
    return ExtractedMl(metadata, windows)


def extract_ml_from_packet(packet: WrappedRedvoxPacketM) -> Optional[ExtractedMl]:
    """
    Extract ML parameters from an event stream.
    :param packet: The packet to extract ML parameters from.
    :return: The extracted ML parameters or None.
    """
    stream: Optional[EventStream] = find_ml_event_stream(packet)
    if stream is None:
        return None
    return extract_ml_from_event_stream(stream)


def extract_ml_from_file(file_path: str) -> Optional[ExtractedMl]:
    """
    Extract ML parameters from a file.
    :param file_path: The path of the packet to extract ML parameters from.
    :return: The extracted ML parameters or None.
    """
    packet: WrappedRedvoxPacketM = WrappedRedvoxPacketM.from_compressed_path(file_path)
    return extract_ml_from_packet(packet)



"""
Read Redvox data from a single directory
Data files can be either API 900 or API 1000 data formats
"""
from typing import List, Optional, Set
from datetime import timedelta, datetime
import multiprocessing
import multiprocessing.pool

import pyarrow as pa
import psutil

import redvox.settings as settings
import redvox.api1000.proto.redvox_api_m_pb2 as api_m
import redvox.common.date_time_utils as dtu
from redvox.common import io, api_conversions as ac
from redvox.common.parallel_utils import maybe_parallel_map
from redvox.common.station import Station, STATION_ID_LENGTH
from redvox.common.reader_session_model import ModelsContainer
from redvox.common.session_model import SessionModel
from redvox.common.errors import RedVoxExceptions, RedVoxError
from redvox.cloud.client import cloud_client
from redvox.cloud.session_model_api import Session
from redvox.cloud.errors import CloudApiError


id_py_stct = pa.struct(
    [
        ("id", pa.string()),
        ("uuid", pa.string()),
        ("start_time", pa.float64()),
    ]
)
meta_py_stct = pa.struct(
    [
        ("api", pa.float64()),
        ("sub_api", pa.float64()),
        ("make", pa.string()),
        ("model", pa.string()),
        ("os", pa.int64()),
        ("os_version", pa.string()),
        ("app", pa.string()),
        ("app_version", pa.string()),
        ("is_private", pa.bool_()),
        ("packet_duration_s", pa.float64()),
        ("station_description", pa.string()),
    ]
)


PERCENT_FREE_MEM_USE = 0.8  # Percentage of total free memory to use when creating stations (1. is 100%)


def zfill_ids(ids_set: Set) -> Set:
    """
    :param ids_set: set of string ids to zero pad to 10 characters long
    :return: the list of ids with zero padded versions added
    """
    s_ids = ids_set.copy()
    for f in ids_set:
        if f and len(f) < STATION_ID_LENGTH:
            s_ids.add(f.zfill(STATION_ID_LENGTH))
    return s_ids


def strip_leading_zero_ids(ids_set: Set) -> Set:
    """
    :param ids_set: set of string ids
    :return: the list of ids with any leading 0's removed
    """
    s_ids = ids_set.copy()
    for f in ids_set:
        s_ids.add(f.lstrip("0"))
    return s_ids


class ApiReader:
    """
    Reads data from api 900 or api 1000 format, converting all data read into RedvoxPacketM for
    ease of comparison and use.

    Properties:
        filter: io.ReadFilter with the station ids, start and end time, start and end time padding, and
        types of files to read

        base_dir: str of the directory containing all the files to read

        structured_dir: bool, if True, the base_dir contains a specific directory structure used by the
        respective api formats.  If False, base_dir only has the data files.  Default False.

        files_index: io.Index of the files that match the filter that are in base_dir

        index_summary: io.IndexSummary of the filtered data

        session_models: ModelContainer for cloud and local session models.

        debug: bool, if True, output additional information during function execution.  Default False.
    """

    def __init__(
        self,
        base_dir: str,
        structured_dir: bool = False,
        read_filter: io.ReadFilter = None,
        debug: bool = False,
        pool: Optional[multiprocessing.pool.Pool] = None,
    ):
        """
        Initialize the ApiReader object

        :param base_dir: directory containing the files to read
        :param structured_dir: if True, base_dir contains a specific directory structure used by the respective
                                api formats.  If False, base_dir only has the data files.  Default False.
        :param read_filter: ReadFilter for the data files, if None, get everything.  Default None
        :param debug: if True, output program warnings/errors during function execution.  Default False.
        """
        _pool: multiprocessing.pool.Pool = multiprocessing.Pool() if pool is None else pool

        if read_filter:
            self.filter: io.ReadFilter = read_filter
            if self.filter.station_ids:
                # add leading 0s to shorter ids
                long_ids = zfill_ids(self.filter.station_ids)
                # remove leading 0s from input ids
                short_ids = strip_leading_zero_ids(self.filter.station_ids)
                self.filter.station_ids = long_ids.union(short_ids)
        else:
            self.filter = io.ReadFilter()
        self.base_dir: str = base_dir
        self.structured_dir: bool = structured_dir
        self.debug: bool = debug
        self.errors: RedVoxExceptions = RedVoxExceptions("APIReader")
        self.session_models: ModelsContainer = ModelsContainer()
        self.files_index: List[io.Index] = self._get_all_files(_pool)
        self.index_summary: io.IndexSummary = io.IndexSummary.from_index(self._flatten_files_index())
        if len(self.files_index) > 0:
            mem_split_factor = len(self.files_index) if settings.is_parallelism_enabled() else 1
            self.chunk_limit = psutil.virtual_memory().available * PERCENT_FREE_MEM_USE / mem_split_factor
            max_file_size = max([fe.decompressed_file_size_bytes for fi in self.files_index for fe in fi.entries])
            total_est_size = max_file_size * sum([len(fi.entries) for fi in self.files_index])
            if max_file_size > self.chunk_limit:
                raise MemoryError(
                    f"System requires {max_file_size} bytes of memory to process a file but only has "
                    f"{self.chunk_limit} available.  Please free or add more RAM."
                )
            elif total_est_size / mem_split_factor > self.chunk_limit:
                raise MemoryError(
                    f"{total_est_size} of data requested, but only {self.chunk_limit} available; "
                    f"please reduce the amount of data you are requesting."
                )
            if debug:
                if mem_split_factor == 1:
                    print(
                        f"{len(self.files_index)} stations have {int(self.chunk_limit)} "
                        f"bytes for loading files in memory."
                    )
                else:
                    print(
                        f"{mem_split_factor} stations each have "
                        f"{int(self.chunk_limit)} bytes for loading files in memory."
                    )
        else:
            self.chunk_limit = 0

        if debug:
            self.errors.print()

        if pool is None:
            _pool.close()

    def _flatten_files_index(self):
        """
        :return: flattened version of files_index
        """
        result = io.Index()
        for i in self.files_index:
            result.append(iter(i.entries))
        return result

    def _get_cloud_models(self, ids: List[str]):
        """
        saves the cloud models from the server that match the list of ids given to the ApiReader's session_models.
        Refer to https://github.com/RedVoxInc/redvox-python-sdk/tree/master/docs/python_sdk/cloud#authenticating-with-redvox-cloud-services
        for more information about how to store credentials on your machine.

        :param ids: station ids to get models for
        """
        try:
            with cloud_client() as client:
                self.session_models.search_cloud_session(
                    id_uuids=ids,
                    owner=client.redvox_config.username,
                    start_ts=int(dtu.datetime_to_epoch_microseconds_utc(self.filter.start_dt))
                    if self.filter.start_dt
                    else None,
                    end_ts=int(dtu.datetime_to_epoch_microseconds_utc(self.filter.end_dt))
                    if self.filter.end_dt
                    else None,
                    include_public=True,
                )
                if self.session_models.cloud_models is None:
                    self.errors.append(f"Unable to find any cloud sessions for {ids}.  Using local files.")
        except CloudApiError as e:
            self.errors.append(f"Error while connecting to server.  Error message: {e}")
        except Exception as e:
            self.errors.append(f"An error occurred.  Error message: {e}")

    def _reset_index(self, model: Session) -> io.Index:
        """
        reset the filter used to get files, then get the updated list of files

        :param model: model to use to reset filter
        :return: updated index of files
        """
        insufficient_str = ""
        # reset the filter used to get files
        new_filter = (
            io.ReadFilter()
            .with_extensions(self.filter.extensions)
            .with_api_versions(self.filter.api_versions)
            .with_station_ids(zfill_ids({model.id}))
            .with_start_dt_buf(dtu.timedelta(seconds=0))
            .with_end_dt_buf(dtu.timedelta(seconds=0))
        )
        # update the start and end times for the filter by the mean offset and the packet duration
        off_adjustment = abs(model.timing.mean_off)
        delta_fix = timedelta(microseconds=(off_adjustment + model.packet_dur))
        if self.filter.start_dt is not None:
            if timedelta(microseconds=off_adjustment) > self.filter.start_dt_buf and model.timing.mean_off < 0:
                insufficient_str += "start "
            new_filter.with_start_dt(self.filter.start_dt - delta_fix)
        if self.filter.end_dt is not None:
            if timedelta(microseconds=off_adjustment) > self.filter.end_dt_buf and model.timing.mean_off > 0:
                insufficient_str += "end"
            new_filter.with_end_dt(self.filter.end_dt + delta_fix)

        if len(insufficient_str) > 0:
            self.errors.append(f"Required more data for {model.id} at: {insufficient_str}")
        return self._apply_filter(new_filter)

    def _get_all_files(self, pool: Optional[multiprocessing.pool.Pool] = None) -> List[io.Index]:
        """
        get all files in the base dir of the ApiReader

        :return: index with all the files that match the filter
        """
        _pool: multiprocessing.pool.Pool = multiprocessing.Pool() if pool is None else pool
        index: List[io.Index] = []
        # this guarantees that all ids we search for are valid
        all_index = self._apply_filter(pool=_pool)
        all_index_ids = all_index.summarize().station_ids()
        # get models using the cloud to correct timing
        self._get_cloud_models(all_index_ids)
        resp_ids = self.session_models.list_ids()

        for station_id in all_index_ids:
            # if start and end are both not defined, just use what we got
            if self.filter.start_dt is None and self.filter.end_dt is None:
                checked_index = all_index.get_index_for_station_id(station_id)
            # if we need to update the start or end, use the first session model from cloud if it exists
            elif station_id in resp_ids:
                checked_index = self._reset_index(self.session_models.get_model_by_partial_key(station_id))
            # if no models from cloud, use the data available to update start and end of index
            else:
                id_index = all_index.get_index_for_station_id(station_id)
                if len(id_index.entries) < 1:
                    continue  # if nothing found, just skip the index
                # attempt to make a session model using local data.  if failure, use what we got initially.
                try:
                    stats = SessionModel().create_from_stream(self.read_files_in_index(id_index))
                    checked_index = self._reset_index(stats.cloud_session)
                    self.session_models.add_local_session(stats)
                except (RedVoxError, Exception):
                    checked_index = id_index

            # add the updated list of files to the index
            if len(checked_index.entries) > 0:
                index.append(checked_index)

        if pool is None:
            _pool.close()

        if len(all_index_ids) > 0:
            self.filter.station_ids = set(all_index_ids)

        return index

    def _apply_filter(
        self,
        reader_filter: Optional[io.ReadFilter] = None,
        pool: Optional[multiprocessing.pool.Pool] = None,
    ) -> io.Index:
        """
        apply the filter of the reader, or another filter if specified

        :param reader_filter: optional filter; if None, use the reader's filter, default None
        :return: index of the filtered files
        """
        _pool: multiprocessing.pool.Pool = multiprocessing.Pool() if pool is None else pool
        if not reader_filter:
            reader_filter = self.filter
        if self.structured_dir:
            index = io.index_structured(self.base_dir, reader_filter, pool=_pool)
        else:
            index = io.index_unstructured(self.base_dir, reader_filter, pool=_pool)
        if pool is None:
            _pool.close()
        return index

    def _redo_index(self, station_ids: set, new_start: datetime, new_end: datetime) -> Optional[io.Index]:
        """
        Redo the index for files using new start and end dates.  removes any buffer time at the start and end of the
        new query.  Returns the updated index or None

        :param station_ids: set of ids to get
        :param new_start: new start time to get data from
        :param new_end: new end time to get data from
        :return: Updated index or None
        """
        new_index = self._apply_filter(
            io.ReadFilter()
            .with_start_dt(new_start)
            .with_end_dt(new_end)
            .with_extensions(self.filter.extensions)
            .with_api_versions(self.filter.api_versions)
            .with_station_ids(station_ids)
            .with_start_dt_buf(timedelta(seconds=0))
            .with_end_dt_buf(timedelta(seconds=0))
        )
        if len(new_index.entries) > 0:
            return new_index
        return None

    def _split_workload(self, findex: io.Index) -> List[io.Index]:
        """
        takes an index and splits it into chunks based on a size limit
        while running_total + next_file_size < limit, adds files to a chunk (Index)
        if limit is exceeded, adds the chunk and puts the next file into a new chunk

        :param findex: index of files to split
        :return: list of Index to process
        """
        packet_list = []
        chunk_queue = 0
        chunk_list = []
        for f in findex.entries:
            chunk_queue += f.decompressed_file_size_bytes
            if chunk_queue > self.chunk_limit:
                packet_list.append(io.Index(chunk_list))
                chunk_queue = 0
                chunk_list = []
            chunk_list.append(f)
        packet_list.append(io.Index(chunk_list))
        return packet_list

    @staticmethod
    def read_files_in_index(indexf: io.Index) -> List[api_m.RedvoxPacketM]:
        """
        read all the files in the index

        :return: list of RedvoxPacketM, converted from API 900 if necessary
        """
        result: List[api_m.RedvoxPacketM] = []

        # Iterate over the API 900 packets in a memory efficient way
        # and convert to API 1000
        # noinspection PyTypeChecker
        for packet_900 in indexf.stream_raw(io.ReadFilter.empty().with_api_versions({io.ApiVersion.API_900})):
            # noinspection Mypy
            result.append(ac.convert_api_900_to_1000_raw(packet_900))

        # Grab the API 1000 packets
        # noinspection PyTypeChecker
        for packet in indexf.stream_raw(io.ReadFilter.empty().with_api_versions({io.ApiVersion.API_1000})):
            # noinspection Mypy
            result.append(packet)

        return result

    # noinspection PyTypeChecker
    def read_files_by_id(self, station_id: str) -> Optional[List[api_m.RedvoxPacketM]]:
        """
        :param station_id: the id to filter on
        :return: the list of packets with the requested id, or None if the id can't be found
        """
        result: List[api_m.RedvoxPacketM] = []

        # Iterate over the API 900 packets in a memory efficient way
        # and convert to API 1000
        for packet_900 in self._flatten_files_index().stream_raw(
            io.ReadFilter.empty().with_api_versions({io.ApiVersion.API_900}).with_station_ids({station_id})
        ):
            # noinspection Mypy
            result.append(ac.convert_api_900_to_1000_raw(packet_900))

        # Grab the API 1000 packets
        for packet in self._flatten_files_index().stream_raw(
            io.ReadFilter.empty().with_api_versions({io.ApiVersion.API_1000}).with_station_ids({station_id})
        ):
            # noinspection Mypy
            result.append(packet)

        if len(result) == 0:
            return None

        return result

    def _station_by_index(self, findex: io.Index) -> Station:
        """
        :param findex: index with files to build a station with
        :return: Station built from files in findex
        """
        return Station.create_from_packets(self.read_files_in_index(findex))

    def get_stations(self, pool: Optional[multiprocessing.pool.Pool] = None) -> List[Station]:
        """
        :param pool: optional multiprocessing pool
        :return: List of all stations in the ApiReader
        """
        return list(maybe_parallel_map(pool, self._station_by_index, iter(self.files_index), chunk_size=1))

    def get_station_by_id(self, get_id: str) -> Optional[List[Station]]:
        """
        :param get_id: the id to filter on
        :return: list of all stations with the requested id or None if id can't be found
        """
        result = [s for s in self.get_stations() if s.id() == get_id]
        if len(result) < 1:
            return None
        return result



"""
This module contains classes and functions that support SessionModel.
"""
from typing import List, Optional, Tuple, Dict, Union, Callable
from bisect import insort

import numpy as np

import redvox.api1000.proto.redvox_api_m_pb2 as api_m
from redvox.api1000.wrapped_redvox_packet.sensors.location import LocationProvider
from redvox.cloud import session_model_api as sm
from redvox.common.timesync import TimeSync
from redvox.common.offset_model import GPS_LATENCY_MICROS


NUM_BUFFER_POINTS = 3  # number of data points to keep in a buffer
NUM_BUFFER_TS_POINTS = 8  # number of data points to keep in a timesync buffer
NUM_BUFFER_LOC_POINTS = 3  # number of data points to keep in a location buffer
COLUMN_TO_ENUM_FN = {"location_provider": lambda l: LocationProvider(l).name}

# These are used for checking if a field is present or not
_ACCELEROMETER_FIELD_NAME: str = "accelerometer"
_AMBIENT_TEMPERATURE_FIELD_NAME: str = "ambient_temperature"
_AUDIO_FIELD_NAME: str = "audio"
_COMPRESSED_AUDIO_FIELD_NAME: str = "compressed_audio"
_GRAVITY_FIELD_NAME: str = "gravity"
_GYROSCOPE_FIELD_NAME: str = "gyroscope"
_IMAGE_FIELD_NAME: str = "image"
_LIGHT_FIELD_NAME: str = "light"
_LINEAR_ACCELERATION_FIELD_NAME: str = "linear_acceleration"
_LOCATION_FIELD_NAME: str = "location"
_MAGNETOMETER_FIELD_NAME: str = "magnetometer"
_ORIENTATION_FIELD_NAME: str = "orientation"
_PRESSURE_FIELD_NAME: str = "pressure"
_PROXIMITY_FIELD_NAME: str = "proximity"
_RELATIVE_HUMIDITY_FIELD_NAME: str = "relative_humidity"
_ROTATION_VECTOR_FIELD_NAME: str = "rotation_vector"
_VELOCITY_FIELD_NAME: str = "velocity"
_HEALTH_FIELD_NAME: str = "health"


Sensor = Union[
    api_m.RedvoxPacketM.Sensors.Xyz,
    api_m.RedvoxPacketM.Sensors.Single,
    api_m.RedvoxPacketM.Sensors.Audio,
    api_m.RedvoxPacketM.Sensors.Image,
    api_m.RedvoxPacketM.Sensors.Location,
    api_m.RedvoxPacketM.Sensors.CompressedAudio,
    api_m.RedvoxPacketM.StationInformation.StationMetrics,
]

__SENSOR_NAME_TO_SENSOR_FN: Dict[
    str,
    Optional[
        Callable[
            [api_m.RedvoxPacketM],
            Union[Sensor],
        ]
    ],
] = {
    "unknown": None,
    _HEALTH_FIELD_NAME: lambda packet: packet.station_information.station_metrics,
    _ACCELEROMETER_FIELD_NAME: lambda packet: packet.sensors.accelerometer,
    _AMBIENT_TEMPERATURE_FIELD_NAME: lambda packet: packet.sensors.ambient_temperature,
    _AUDIO_FIELD_NAME: lambda packet: packet.sensors.audio,
    _COMPRESSED_AUDIO_FIELD_NAME: lambda packet: packet.sensors.compressed_audio,
    _GRAVITY_FIELD_NAME: lambda packet: packet.sensors.gravity,
    _GYROSCOPE_FIELD_NAME: lambda packet: packet.sensors.gyroscope,
    _IMAGE_FIELD_NAME: lambda packet: packet.sensors.image,
    _LIGHT_FIELD_NAME: lambda packet: packet.sensors.light,
    _LINEAR_ACCELERATION_FIELD_NAME: lambda packet: packet.sensors.linear_acceleration,
    _LOCATION_FIELD_NAME: lambda packet: packet.sensors.location,
    _MAGNETOMETER_FIELD_NAME: lambda packet: packet.sensors.magnetometer,
    _ORIENTATION_FIELD_NAME: lambda packet: packet.sensors.orientation,
    _PRESSURE_FIELD_NAME: lambda packet: packet.sensors.pressure,
    _PROXIMITY_FIELD_NAME: lambda packet: packet.sensors.proximity,
    _RELATIVE_HUMIDITY_FIELD_NAME: lambda packet: packet.sensors.relative_humidity,
    _ROTATION_VECTOR_FIELD_NAME: lambda packet: packet.sensors.rotation_vector,
    _VELOCITY_FIELD_NAME: lambda packet: packet.sensors.velocity,
}


def _get_sensor_for_data_extraction(sensor_name: str, packet: api_m.RedvoxPacketM) -> Optional[Sensor]:
    """
    :param sensor_name: name of sensor to return
    :param packet: the data packet to get the sensor from
    :return: Sensor that matches the sensor_name or None if that Sensor doesn't exist
    """
    sensor_fn: Optional[Callable[[api_m.RedvoxPacketM], Sensor]] = __SENSOR_NAME_TO_SENSOR_FN[sensor_name]
    if (sensor_name == _HEALTH_FIELD_NAME or _has_sensor(packet, sensor_name)) and sensor_fn is not None:
        return sensor_fn(packet)


def _get_mean_sample_rate_from_sensor(sensor: Sensor) -> float:
    """
    :param sensor: Sensor to get data from
    :return: mean sample rate of the sensor or np.nan if sample rate doesn't exist
    """
    return sensor.timestamps.mean_sample_rate if int(sensor.timestamps.timestamp_statistics.count) > 1 else np.nan


def _has_sensor(data: Union[api_m.RedvoxPacketM, api_m.RedvoxPacketM.Sensors], field_name: str) -> bool:
    """
    Returns true if the given packet or sensors instance contains the valid sensor.

    :param data: Either a packet or a packet's sensors message.
    :param field_name: The name of the sensor being checked.
    :return: True if the sensor exists, False otherwise.
    """
    if isinstance(data, api_m.RedvoxPacketM):
        # noinspection Mypy,PyTypeChecker
        return data.sensors.HasField(field_name)

    if isinstance(data, api_m.RedvoxPacketM.Sensors):
        # noinspection Mypy,PyTypeChecker
        return data.HasField(field_name)

    return False


def get_all_sensors_in_packet(packet: api_m.RedvoxPacketM) -> List[Tuple[str, str, float]]:
    """
    :param packet: packet to check
    :return: list of all sensors as tuple of name, description, and mean sample rate in the packet
    """
    result: List[Tuple] = []
    for s in [_AUDIO_FIELD_NAME, _COMPRESSED_AUDIO_FIELD_NAME]:
        if _has_sensor(packet, s):
            sensor = _get_sensor_for_data_extraction(s, packet)
            result.append((s, sensor.sensor_description, sensor.sample_rate))
    for s in [
        _PRESSURE_FIELD_NAME,
        _LOCATION_FIELD_NAME,
        _ACCELEROMETER_FIELD_NAME,
        _AMBIENT_TEMPERATURE_FIELD_NAME,
        _GRAVITY_FIELD_NAME,
        _GYROSCOPE_FIELD_NAME,
        _IMAGE_FIELD_NAME,
        _LIGHT_FIELD_NAME,
        _LINEAR_ACCELERATION_FIELD_NAME,
        _MAGNETOMETER_FIELD_NAME,
        _ORIENTATION_FIELD_NAME,
        _PROXIMITY_FIELD_NAME,
        _RELATIVE_HUMIDITY_FIELD_NAME,
        _ROTATION_VECTOR_FIELD_NAME,
        _VELOCITY_FIELD_NAME,
    ]:
        if _has_sensor(packet, s):
            sensor = _get_sensor_for_data_extraction(s, packet)
            result.append((s, sensor.sensor_description, sensor.timestamps.mean_sample_rate))
    if packet.station_information.HasField("station_metrics"):
        result.insert(
            2,
            (
                _HEALTH_FIELD_NAME,
                "station_metrics",
                packet.station_information.station_metrics.timestamps.mean_sample_rate,
            ),
        )
    return result


def __ordered_insert(buffer: List, value: Tuple):
    """
    inserts the value into the buffer using the timestamp as the key

    :param value: value to add.  Must include a timestamp and the same data type as the other buffer elements
    """
    if len(buffer) < 1:
        buffer.append(value)
    else:
        insort(buffer, value)


def add_to_fst_buffer(buffer: List, buf_max_size: int, timestamp: float, value):
    """
    Add a value into the first buffer.

    * If the buffer is not full, the value is added automatically
    * If the buffer is full, the value is only added if it comes before the last element.
    * Ignores adding any duplicate values

    :param buffer: the buffer to add the value to
    :param buf_max_size: the maximum size of the buffer
    :param timestamp: timestamp in microseconds since epoch UTC to add.
    :param value: value to add.  Must be the same type of data as the other elements in the queue.
    """
    if (len(buffer) < buf_max_size or timestamp < buffer[-1][0]) and timestamp not in [n[0] for n in buffer]:
        __ordered_insert(buffer, (timestamp, value))
        while len(buffer) > buf_max_size:
            buffer.pop()


def add_to_lst_buffer(buffer: List, buf_max_size: int, timestamp: float, value):
    """
    Add a value into the last buffer.

    * If the buffer is not full, the value is added automatically
    * If the buffer is full, the value is only added if it comes after the first element.
    * Ignores adding any duplicate values

    :param buffer: the buffer to add the value to
    :param buf_max_size: the maximum size of the buffer
    :param timestamp: timestamp in microseconds since epoch UTC to add.
    :param value: value to add.  Must be the same type of data as the other elements in the queue.
    """
    if (len(buffer) < buf_max_size or timestamp > buffer[0][0]) and timestamp not in [n[0] for n in buffer]:
        __ordered_insert(buffer, (timestamp, value))
        while len(buffer) > buf_max_size:
            buffer.pop(0)


def get_local_timesync(packet: api_m.RedvoxPacketM) -> Tuple:
    """
    The returning tuple looks like:

    (start_timestamp, end_timestamp, num_exchanges, best_latency, best_offset, list of TimeSyncData)

    num_exchanges, best_latency and best_offset default to 0.  list of TimeSyncData defaults to empty list.

    :param packet: packet to get timesync data from
    :return: Tuple with timing data from packet
    """
    ts = TimeSync().from_raw_packets([packet])
    if ts.num_tri_messages() > 0:
        _ts_latencies = ts.latencies().flatten()
        _ts_offsets = ts.offsets().flatten()
        _ts_timestamps = ts.get_device_exchanges_timestamps()
        # add data to the buffers
        _ts_data = [
            sm.TimeSyncData(_ts_timestamps[i], _ts_latencies[i], _ts_offsets[i]) for i in range(len(_ts_timestamps))
        ]
        return (
            ts.data_start_timestamp(),
            ts.data_end_timestamp(),
            ts.num_tri_messages(),
            ts.best_latency(),
            ts.best_offset(),
            _ts_data,
        )
    return (
        ts.data_start_timestamp(),
        ts.data_end_timestamp(),
        0.0,
        0.0,
        0.0,
        [],
    )


def get_gps_timing(packet: api_m.RedvoxPacketM) -> Tuple:
    """
    The returning tuple looks like:

    (start_timestamp, end_timestamp, num_samples, best_latency, best_offset, list of TimeSyncData)

    num_samples, best_latency and best_offset default to 0.  list of TimeSyncData defaults to empty list.

    :param packet: packet to get gps timesync data from
    :return: Tuple with GPS timing data from packet
    """
    loc = packet.sensors.location
    num_pts = int(loc.timestamps.timestamp_statistics.count)
    if num_pts > 0:
        gps_timestamps = np.array(loc.timestamps_gps.timestamps)
        unique_gps = np.unique(gps_timestamps)
        if len(unique_gps) != len(gps_timestamps) and not np.isnan(unique_gps):
            keep_gps = []
            for n in unique_gps:
                instances = np.argwhere(gps_timestamps == n)
                if len(instances[0]) == 1:
                    keep_gps.append(instances[0][0])
            if len(keep_gps) < 1:
                return 0, 0, 0, 0, 0, []
            gps_offsets = (
                np.array([gps_timestamps[i] - loc.timestamps.timestamps[i] for i in keep_gps]) + GPS_LATENCY_MICROS
            )
            timestamps = [loc.timestamps.timestamps[i] for i in keep_gps]
            start_timestamp = gps_timestamps[min(keep_gps)]
            end_timestamp = gps_timestamps[max(keep_gps)]
        else:
            timestamps = loc.timestamps.timestamps
            gps_offsets = gps_timestamps - np.array(timestamps) + GPS_LATENCY_MICROS
            start_timestamp = gps_timestamps[0]
            end_timestamp = gps_timestamps[-1]
        _ts_data = [sm.TimeSyncData(timestamps[i], GPS_LATENCY_MICROS, gps_offsets[i]) for i in range(len(timestamps))]
    else:
        if loc.last_best_location is not None:
            bl = loc.last_best_location
        elif loc.overall_best_location is not None:
            bl = loc.overall_best_location
        else:
            return 0, 0, 0, 0, 0, []
        gps_offsets = bl.latitude_longitude_timestamp.gps - bl.latitude_longitude_timestamp.mach + GPS_LATENCY_MICROS
        start_timestamp = bl.latitude_longitude_timestamp.gps
        timestamps = bl.latitude_longitude_timestamp.mach
        end_timestamp = start_timestamp
        num_pts = 1
        _ts_data = [sm.TimeSyncData(timestamps, GPS_LATENCY_MICROS, gps_offsets)]
    best_offset = np.mean(gps_offsets)
    if np.isnan(best_offset):
        return 0, 0, 0, 0, 0, []
    return start_timestamp, end_timestamp, num_pts, GPS_LATENCY_MICROS, best_offset, _ts_data


def add_to_welford(value: float, welford: Optional[sm.WelfordAggregator] = None) -> sm.WelfordAggregator:
    """
    adds the value to the welford, then returns the updated object.

    If welford is None, creates a new WelfordAggregator object and returns it.

    :param value: the value to add
    :param welford: optional WelfordAggregator object to update.  if not given, will make a new one.  Default None
    :return: updated or new WelfordAggregator object
    """
    if welford is None:
        return sm.WelfordAggregator(0.0, value, 1)
    welford.cnt += 1
    delta = value - welford.mean
    welford.mean += delta / float(welford.cnt)
    delta2 = value - welford.mean
    welford.m2 += delta * delta2
    return welford


def add_to_stats(value: float, stats: Optional[sm.Stats] = None) -> sm.Stats:
    """
    adds the value to the stats, then returns the updated object.

    If stats is None, creates a new Stats object and returns it.

    :param value: the value to add
    :param stats: optional Stats object to update.  if not given, will make a new one.  Default None
    :return: updated or new Stats object
    """
    if stats is None:
        return sm.Stats(value, value, add_to_welford(value))
    if value < stats.min:
        stats.min = value
    if value > stats.max:
        stats.max = value
    add_to_welford(value, stats.welford)
    return stats


def get_location_data(packet: api_m.RedvoxPacketM) -> List[Tuple[str, float, float, float, float]]:
    """
    :param packet: packet to get location data from
    :return: List of location data as a tuples from the packet
    """
    locations = []
    loc = packet.sensors.location
    lat = lon = alt = ts = 0.0
    source = "UNKNOWN"
    num_pts = int(loc.timestamps.timestamp_statistics.count)
    # check for actual location values
    if len(loc.location_providers) < 1 and num_pts > 0:
        lat = loc.latitude_samples.value_statistics.mean
        lon = loc.longitude_samples.value_statistics.mean
        alt = loc.altitude_samples.value_statistics.mean
        ts = loc.timestamps.timestamp_statistics.mean
    elif (
        num_pts > 0
        and loc.latitude_samples.value_statistics.count > 0
        and loc.longitude_samples.value_statistics.count > 0
        and loc.altitude_samples.value_statistics.count > 0
        and num_pts == loc.latitude_samples.value_statistics.count
        and num_pts == loc.altitude_samples.value_statistics.count
        and num_pts == loc.longitude_samples.value_statistics.count
    ):
        lats = loc.latitude_samples.values
        lons = loc.longitude_samples.values
        alts = loc.altitude_samples.values
        tstp = loc.timestamps.timestamps
        for i in range(num_pts):
            source = (
                "UNKNOWN"
                if len(loc.location_providers) != num_pts
                else COLUMN_TO_ENUM_FN["location_provider"](loc.location_providers[i])
            )
            locations.append((source, lats[i], lons[i], alts[i], tstp[i]))
        # set a special flag for later, so we don't add an extra location value
        source = None
    elif loc.last_best_location is not None:
        ts = loc.last_best_location.latitude_longitude_timestamp.mach
        source = COLUMN_TO_ENUM_FN["location_provider"](loc.last_best_location.location_provider)
        lat = loc.last_best_location.latitude
        lon = loc.last_best_location.longitude
        alt = loc.last_best_location.altitude
    elif loc.overall_best_location is not None:
        ts = loc.overall_best_location.latitude_longitude_timestamp.mach
        source = COLUMN_TO_ENUM_FN["location_provider"](loc.overall_best_location.location_provider)
        lat = loc.overall_best_location.latitude
        lon = loc.overall_best_location.longitude
        alt = loc.overall_best_location.altitude
    # source is not None if we got only one location through non-usual methods
    if source is not None:
        locations.append((source, lat, lon, alt, ts))
    return locations


def get_dynamic_data(packet: api_m.RedvoxPacketM) -> Dict:
    """
    :param packet: packet to get data from
    :return: Dictionary of all dynamic session data from the packet
    """
    location = get_location_data(packet)
    battery = packet.station_information.station_metrics.battery.value_statistics.mean
    temperature = packet.station_information.station_metrics.temperature.value_statistics.mean
    return {"location": location, "battery": battery, "temperature": temperature}


def add_to_location(
    lat: float, lon: float, alt: float, timestamp: float, loc_stat: Optional[sm.LocationStat] = None
) -> sm.LocationStat:
    """
    update a LocationStat object with the location, or make a new one

    :param lat: latitude in degrees
    :param lon: longitude in degrees
    :param alt: altitude in meters
    :param timestamp: timestamp in microseconds from epoch UTC
    :param loc_stat: optional LocationStat object to update.  if not given, will make a new one.  Default None
    :return: updated or new LocationStat object
    """
    if loc_stat is None:
        fst_lst = sm.FirstLastBufLocation([], NUM_BUFFER_LOC_POINTS, [], NUM_BUFFER_LOC_POINTS)
        add_to_fst_buffer(fst_lst.fst, fst_lst.fst_max_size, timestamp, sm.Location(lat, lon, alt))
        add_to_lst_buffer(fst_lst.lst, fst_lst.lst_max_size, timestamp, sm.Location(lat, lon, alt))
        return sm.LocationStat(fst_lst, add_to_stats(lat), add_to_stats(lon), add_to_stats(alt))
    add_to_fst_buffer(loc_stat.fst_lst.fst, loc_stat.fst_lst.fst_max_size, timestamp, sm.Location(lat, lon, alt))
    add_to_lst_buffer(loc_stat.fst_lst.lst, loc_stat.fst_lst.lst_max_size, timestamp, sm.Location(lat, lon, alt))
    loc_stat.lat = add_to_stats(lat, loc_stat.lat)
    loc_stat.lng = add_to_stats(lon, loc_stat.lng)
    loc_stat.alt = add_to_stats(alt, loc_stat.alt)
    return loc_stat


def add_location_data(
    data: List[Tuple[str, float, float, float, float]], loc_dict: Optional[Dict[str, sm.LocationStat]] = None
) -> Dict[str, sm.LocationStat]:
    """
    update a dictionary of LocationStat or make a new dictionary

    :param data: the data to add
    :param loc_dict: the location dictionary to update
    :return: the updated or new location dictionary
    """
    if loc_dict is None:
        loc_dict = {}
    for s in data:
        loc_dict[s[0]] = add_to_location(s[1], s[2], s[3], s[4], loc_dict[s[0]] if s[0] in loc_dict.keys() else None)
    return loc_dict



"""
This module provides classes to organize events recorded on a station.
It will ignore machine learning events.
"""
from typing import List, Optional, Dict, Union
from dataclasses import dataclass, field
from pathlib import Path
import enum
import os
import re

import numpy as np
from dataclasses_json import dataclass_json

from redvox.api1000.common.mapping import Mapping
from redvox.api1000.proto.redvox_api_m_pb2 import RedvoxPacketM
from redvox.api1000.wrapped_redvox_packet import event_streams as es
from redvox.api1000.wrapped_redvox_packet import ml
from redvox.api1000.wrapped_redvox_packet.wrapped_packet import WrappedRedvoxPacketM
from redvox.common.errors import RedVoxExceptions
from redvox.common import offset_model as om
from redvox.common.io import FileSystemWriter as Fsw, FileSystemSaveMode, json_file_to_dict
import redvox.common.event_stream_io as io


class EventDataTypes(enum.Enum):
    """
    Enumeration of data types for event data
    """

    STRING = 0  # string data
    NUMERIC = 1  # numeric data
    BOOLEAN = 2  # boolean data
    BYTE = 3  # bytes data

    @staticmethod
    def types_list() -> List["EventDataTypes"]:
        """
        :return: the values of EventDataTypes as a list in order of: STRING, NUMERIC, BOOLEAN, BYTE
        """
        return [EventDataTypes.STRING, EventDataTypes.NUMERIC, EventDataTypes.BOOLEAN, EventDataTypes.BYTE]


def get_empty_event_data_dict() -> dict:
    """
    :return: an empty data dictionary for event data
    """
    return {EventDataTypes.STRING: {}, EventDataTypes.NUMERIC: {}, EventDataTypes.BOOLEAN: {}, EventDataTypes.BYTE: {}}


class Event:
    """
    stores event data from Redvox Api1000 packets

    ALL timestamps in microseconds since epoch UTC unless otherwise stated
    """

    def __init__(
        self,
        timestamp: float,
        name: str = "event",
        data: Optional[Dict[EventDataTypes, dict]] = None,
        save_mode: FileSystemSaveMode = FileSystemSaveMode.MEM,
        base_dir: str = ".",
    ):
        """
        initialize Event

        :param timestamp: timestamp when Event occurred in microseconds since epoch UTC
        :param name: name of the Event.  Default "event"
        :param data: a structured dictionary of the data.  Dictionary must look like:
                    {EventDataTypes.STRING: {s_values}, EventDataTypes.NUMERIC: {n_values},
                    EventDataTypes.BOOLEAN: {o_values}, EventDataTypes.BYTE: {b_values}}
                    where {*_values} is a dictionary of string: data and can be empty.  Default None
        :param save_mode: FileSystemSaveMode that determines how data is saved.
                            Default FileSystemSaveMode.MEM (use RAM).  Other options are DISK (save to directory)
                            and TEMP (save to temporary directory)
        :param base_dir: the location of the parquet file that holds the data.  Not used if save_data is False.
                            Default current directory (".")
        """
        self.name = name
        self.metadata = {}
        self._errors = RedVoxExceptions("Event")
        self._fs_writer = Fsw(f"event_{name}", "json", base_dir, save_mode)
        self._timestamp = timestamp
        self._uncorrected_timestamp = timestamp
        self._data = get_empty_event_data_dict() if data is None else data

    def __repr__(self):
        return (
            f"name: {self.name}, "
            f"timestamp: {self._timestamp}, "
            f"uncorrected_timestamp: {self._uncorrected_timestamp}, "
            f"schema: {self.get_schema()}, "
            f"save_mode: {self._fs_writer.save_mode()}"
        )

    def __str__(self):
        return (
            f"name: {self.name}, "
            f"timestamp: {self._timestamp}, "
            f"uncorrected_timestamp: {self._uncorrected_timestamp}, "
            f"schema: {self.__schema_as_str()}"
        )

    def as_dict(self) -> dict:
        """
        :return: EventStream as a dictionary
        """
        return {
            "name": self.name,
            "timestamp": self._timestamp,
            "uncorrected_timestamp": self._uncorrected_timestamp,
            "metadata": self.metadata,
            "data": self.__data_as_dict(),
            "errors": self._errors.as_dict(),
        }

    def __data_as_dict(self) -> dict:
        return {
            EventDataTypes.STRING.name: self.get_string_values(),
            EventDataTypes.NUMERIC.name: self.get_numeric_values(),
            EventDataTypes.BOOLEAN.name: self.get_boolean_values(),
            EventDataTypes.BYTE.name: self.get_byte_values(),
        }

    def __schema_as_str(self) -> str:
        result = ""
        for f in self._data.keys():
            result += f"{f.name}: {list(self._data[f].keys())}"
            if f != EventDataTypes.BYTE:
                result += ", "
        return result

    @staticmethod
    def __get_items(payload: Mapping[str]):
        return payload.get_metadata().items()

    @staticmethod
    def __get_items_raw(payload):
        return payload.items()

    @staticmethod
    def __get_keys(ptype: str, payload: Mapping[str]):
        return ptype, payload.get_metadata().keys()

    @staticmethod
    def __get_keys_raw(ptype: str, payload):
        return ptype, payload.keys()

    @staticmethod
    def __get_data_from_event(event: es.Event):
        """
        load data from an Event;
        gets data in order of: string, numeric, boolean, byte

        :param event: event to load data from
        """
        return map(
            Event.__get_items,
            [
                event.get_string_payload(),
                event.get_numeric_payload(),
                event.get_boolean_payload(),
                event.get_byte_payload(),
            ],
        )

    @staticmethod
    def __get_data_from_event_raw(event: RedvoxPacketM.EventStream.Event):
        """
        load data from an Event;
        gets data in order of: string, numeric, boolean, byte

        :param event: event to load data from
        """
        return map(
            Event.__get_items_raw,
            [event.string_payload, event.numeric_payload, event.boolean_payload, event.byte_payload],
        )

    def _set_data(self, data: iter):
        """
        sets the data of the Event

        :param data: an iterable of data to insert
        """
        for g, h in map(lambda l, p: (l, p), data, EventDataTypes.types_list()):
            for k, v in g:
                self._data[h][k] = v

    def read_event(self, event: es.Event) -> "Event":
        """
        read the payloads of a Redvox Event, separate the data by payload type, then add it to the SDK Event

        :param event: event to process
        :return: updated self
        """
        self.name = event.get_description()
        self._fs_writer.file_name = f"event_{self.name}"
        self.metadata = event.get_metadata()
        self._set_data(self.__get_data_from_event(event))
        return self

    def read_raw(self, event: RedvoxPacketM.EventStream.Event) -> "Event":
        """
        read the contents of a Redvox Api1000 protobuf stream

        :param event: the protobuf stream to read
        """
        self.name = event.description
        self._fs_writer.file_name = f"event_{self.name}"
        self.metadata = dict(event.metadata)
        self._set_data(self.__get_data_from_event_raw(event))
        return self

    def get_string_schema(self) -> List[str]:
        """
        :return: the column names of string typed data as a list of strings
        """
        return self.get_schema()[EventDataTypes.STRING]

    def get_numeric_schema(self) -> List[str]:
        """
        :return: the column names of numeric typed data as a list of strings
        """
        return self.get_schema()[EventDataTypes.NUMERIC]

    def get_boolean_schema(self) -> List[str]:
        """
        :return: the column names of boolean typed data as a list of strings
        """
        return self.get_schema()[EventDataTypes.BOOLEAN]

    def get_byte_schema(self) -> List[str]:
        """
        :return: the column names of byte typed data as a list of strings
        """
        return self.get_schema()[EventDataTypes.BYTE]

    def get_schema(self) -> Dict[EventDataTypes, list]:
        """
        :return: the dictionary that summarizes the data names and types
        """
        result = {}
        for f in self._data.keys():
            result[f] = [k for k in self._data[f].keys()]
        return result

    def get_data_keys(self) -> List[str]:
        """
        :return: the keys of the data in the event
        """
        result = []
        for f in self._data.keys():
            result.extend([k for k in self._data[f].keys()])
        return result

    def get_string_values(self) -> dict:
        """
        :return: the string data as a dictionary
        """
        return self._data[EventDataTypes.STRING]

    def get_numeric_values(self) -> dict:
        """
        :return: the numeric data as a dictionary
        """
        return self._data[EventDataTypes.NUMERIC]

    def get_boolean_values(self) -> dict:
        """
        :return: the boolean data as a dictionary
        """
        return self._data[EventDataTypes.BOOLEAN]

    def get_byte_values(self) -> dict:
        """
        :return: the byte data as a dictionary
        """
        return self._data[EventDataTypes.BYTE]

    def get_string_item(self, data_key: str) -> Optional[str]:
        """
        get a string data value with a key matching data_key

        :param data_key: the name of the data value to look for
        :return: string data if it exists, None otherwise
        """
        strs = self.get_string_values()
        for s in strs.keys():
            if s == data_key:
                return strs[s]
        return None

    def get_numeric_item(self, data_key: str) -> Optional[float]:
        """
        get a numeric data value with a key matching data_key

        :param data_key: the name of the data value to look for
        :return: numeric data if it exists, None otherwise
        """
        nums = self.get_numeric_values()
        for s in nums.keys():
            if s == data_key:
                return nums[s]
        return None

    def get_boolean_item(self, data_key: str) -> Optional[bool]:
        """
        get a boolean data value with a key matching data_key

        :param data_key: the name of the data value to look for
        :return: boolean data if it exists, None otherwise
        """
        boos = self.get_boolean_values()
        for s in boos.keys():
            if s == data_key:
                return boos[s]
        return None

    def get_byte_item(self, data_key: str) -> Optional[str]:
        """
        get a byte data value with a key matching data_key

        :param data_key: the name of the data value to look for
        :return: byte data if it exists, None otherwise
        """
        byts = self.get_byte_values()
        for s in byts.keys():
            if s == data_key:
                return byts[s]
        return None

    def get_item(self, data_key: str) -> Union[List[str], str, bool, float]:
        """
        :param data_key: key of data to get
        :return: data with matching data_key or the list of all possible data keys
        """
        for r in [
            self.get_string_item(data_key),
            self.get_numeric_item(data_key),
            self.get_boolean_item(data_key),
            self.get_byte_item(data_key),
        ]:
            if r is not None:
                return r
        return self.get_data_keys()

    def get_classification(self, index: int = 0) -> dict:
        """
        get a classification from an event (anything that ends with "_X" where X is the index value)

        :param index: index of classification, default 0
        :return: dictionary of data
        """
        result = {}
        for s in self._data.keys():
            for b, v in self._data[s].items():
                match = re.search(f"_{index}$", b)
                if match is not None:
                    result[b] = v
        return result

    def get_string_column(self, column_name: str) -> Dict[str, str]:
        """
        note: data points in events are named [column_name]_[X], where [X] is an integer 0 or greater.
        this function will return all string data points with keys that start with [column_name]

        :param column_name: the name of the column of event data to get
        :return: a dictionary of string data
        """
        result = {}
        strs = self.get_string_values()
        for s, v in strs.items():
            match = re.match(f"{column_name}_*", s)
            if match is not None:
                result[s] = v
        return result

    def get_numeric_column(self, column_name: str) -> Dict[str, float]:
        """
        note: data points in events are named [column_name]_[X], where [X] is an integer 0 or greater.
        this function will return all numeric data points with keys that start with [column_name]

        :param column_name: the name of the column of event data to get
        :return: a dictionary of numeric data
        """
        result = {}
        strs = self.get_numeric_values()
        for s, v in strs.items():
            match = re.match(f"{column_name}_*", s)
            if match is not None:
                result[s] = v
        return result

    def get_boolean_column(self, column_name: str) -> Dict[str, str]:
        """
        note: data points in events are named [column_name]_[X], where [X] is an integer 0 or greater.
        this function will return all boolean data points with keys that start with [column_name]

        :param column_name: the name of the column of event data to get
        :return: a dictionary of boolean data
        """
        result = {}
        strs = self.get_boolean_values()
        for s, v in strs.items():
            match = re.match(f"{column_name}_*", s)
            if match is not None:
                result[s] = v
        return result

    def get_byte_column(self, column_name: str) -> Dict[str, str]:
        """
        note: data points in events are named [column_name]_[X], where [X] is an integer 0 or greater.
        this function will return all byte data points with keys that start with [column_name]

        :param column_name: the name of the column of event data to get
        :return: a dictionary of byte data
        """
        result = {}
        strs = self.get_byte_values()
        for s, v in strs.items():
            match = re.match(f"{column_name}_*", s)
            if match is not None:
                result[s] = v
        return result

    def get_timestamp(self) -> float:
        """
        :return: timestamp of the Event
        """
        return self._timestamp

    def get_uncorrected_timestamp(self) -> float:
        """
        :return: uncorrected timestamp of the Event
        """
        return self._uncorrected_timestamp

    def is_timestamp_corrected(self) -> bool:
        """
        :return: if timestamp of Event is updated
        """
        return self._timestamp != self._uncorrected_timestamp

    def update_timestamps(self, offset_model: om.OffsetModel, use_model_function: bool = False):
        """
        updates the timestamp of the Event

        :param offset_model: model used to update the timestamps
        :param use_model_function: if True, use the model's slope function to update the timestamps.
                                    otherwise uses the best offset (model's intercept value).  Default False
        """
        if self.is_timestamp_corrected():
            self._errors.append("Timestamps already corrected!")
        else:
            self._timestamp = offset_model.update_time(self._timestamp, use_model_function)

    def original_timestamps(self, offset_model: om.OffsetModel, use_model_function: bool = False):
        """
        undo the update to the timestamp of the Event

        :param offset_model: model used to update the timestamps
        :param use_model_function: if True, use the model's slope function to update the timestamps.
                                    otherwise uses the best offset (model's intercept value).  Default False
        """
        if not self.is_timestamp_corrected():
            self._errors.append("Timestamps already not corrected!")
        else:
            self._timestamp = offset_model.get_original_time(self._timestamp, use_model_function)

    def default_json_file_name(self) -> str:
        """
        :return: default event json file name (event_[event.name]): note there is no extension
        """
        return f"event_{self.name}"

    def is_save_to_disk(self) -> bool:
        """
        :return: True if sensor will be saved to disk
        """
        return self._fs_writer.is_save_disk()

    def set_save_to_disk(self, save: bool):
        """
        :param save: If True, save to disk
        """
        self._fs_writer.save_to_disk = save

    def set_save_mode(self, save_mode: FileSystemSaveMode):
        """
        set the save mode

        :param save_mode: new save mode
        """
        self._fs_writer.set_save_mode(save_mode)

    def save_mode(self) -> FileSystemSaveMode:
        """
        :return: the save mode
        """
        return self._fs_writer.save_mode()

    def set_file_name(self, new_file: Optional[str] = None):
        """
        * set the pyarrow file name or use the default: event_{Event.name}
        * Do not give an extension

        :param new_file: optional file name to change to; default None (use default name)
        """
        self._fs_writer.file_name = new_file if new_file else f"event_{self.name}"

    def full_file_name(self) -> str:
        """
        :return: full name of file containing the data
        """
        return self._fs_writer.full_name()

    def file_name(self) -> str:
        """
        :return: file name without extension
        """
        return self._fs_writer.file_name

    def set_save_dir(self, new_dir: Optional[str] = None):
        """
        set the pyarrow directory or use the default: "." (current directory)

        :param new_dir: the directory to change to; default None (use current directory)
        """
        self._fs_writer.base_dir = new_dir if new_dir else "."

    def save_dir(self) -> str:
        """
        :return: directory containing parquet files for the sensor
        """
        return self._fs_writer.save_dir()

    def full_path(self) -> str:
        """
        :return: the full path to the data file
        """
        return self._fs_writer.full_path()

    def fs_writer(self) -> Fsw:
        """
        :return: FileSystemWriter object
        """
        return self._fs_writer

    def has_data(self) -> bool:
        """
        :return: True if Event contains at least one data point
        """
        return sum([len(self._data[j].keys()) for j in EventDataTypes.types_list()]) > 0

    def data(self) -> dict:
        """
        :return: the data
        """
        return self._data

    @staticmethod
    def from_json_dict(json_dict: dict) -> "Event":
        """
        :param json_dict: json dictionary to parse
        :return: Event from json dict
        """
        if "timestamp" in json_dict.keys():
            data = get_empty_event_data_dict()
            data[EventDataTypes.STRING] = json_dict["data"]["STRING"]
            data[EventDataTypes.NUMERIC] = json_dict["data"]["NUMERIC"]
            data[EventDataTypes.BOOLEAN] = json_dict["data"]["BOOLEAN"]
            data[EventDataTypes.BYTE] = json_dict["data"]["BYTE"]
            result = Event(json_dict["timestamp"], json_dict["name"], data, FileSystemSaveMode.DISK)
            result.metadata = json_dict["metadata"]
            result._uncorrected_timestamp = json_dict["uncorrected_timestamp"]
            result.set_errors(RedVoxExceptions.from_dict(json_dict["errors"]))
        else:
            result = Event(np.nan, "Empty")
            result.append_error(f"Loading from json dict failed; missing Event timestamp.")
        return result

    @staticmethod
    def from_json_file(file_dir: str, file_name: str) -> "Event":
        """
        :param file_dir: full path to containing directory for the file
        :param file_name: name of file to load data from
        :return: Event from json file
        """
        json_data = json_file_to_dict(os.path.join(file_dir, f"{file_name}"))
        if "timestamp" in json_data.keys():
            data = get_empty_event_data_dict()
            data[EventDataTypes.STRING] = json_data["data"]["STRING"]
            data[EventDataTypes.NUMERIC] = json_data["data"]["NUMERIC"]
            data[EventDataTypes.BOOLEAN] = json_data["data"]["BOOLEAN"]
            data[EventDataTypes.BYTE] = json_data["data"]["BYTE"]
            result = Event(json_data["timestamp"], json_data["name"], data, FileSystemSaveMode.DISK, file_dir)
            result.metadata = json_data["metadata"]
            result._uncorrected_timestamp = json_data["uncorrected_timestamp"]
            result.set_errors(RedVoxExceptions.from_dict(json_data["errors"]))
        else:
            result = Event(np.nan, "Empty")
            result.append_error(f"Loading from {file_name} failed; missing Event timestamp.")
        return result

    def to_json_file(self, file_name: Optional[str] = None) -> Path:
        """
        saves the EventStream as a json file

        :param file_name: the optional base file name.  Do not include a file extension.
                            If None, a default file name is created using this format:
                            event_[event.name].json
        :return: path to json file
        """
        return io.event_to_json_file(self, file_name)

    def errors(self) -> RedVoxExceptions:
        """
        :return: errors of the sensor
        """
        return self._errors

    def set_errors(self, errors: RedVoxExceptions):
        """
        sets the errors of the Sensor

        :param errors: errors to set
        """
        self._errors = errors

    def append_error(self, error: str):
        """
        add an error to the Sensor

        :param error: error to add
        """
        self._errors.append(error)

    def print_errors(self):
        """
        print all errors to screen
        """
        self._errors.print()


@dataclass_json
@dataclass
class EventStream:
    """
    stores multiple events.

    ALL timestamps in microseconds since epoch UTC unless otherwise stated

    Properties:
        name: string; name of the EventStream.  Default "stream"

        events: List[Event]; all events in the stream.  Default empty list

        input_sample_rate: int; audio sample rate.  Default 0

        samples_per_window: int; samples per window of the events.  Default 0

        samples_per_hop: int; samples per hop of the events.  Default 0

        model_version: string; version of the model.  Default "0.0"

        metadata: Dict[str, str]; metadata as dict of strings.  Default empty dict

        debug: boolean; if True, outputs additional information at runtime.  Default False.
    """

    name: str = "stream"
    events: List[Event] = field(default_factory=lambda: [])
    input_sample_rate: int = 0
    samples_per_window: int = 0
    samples_per_hop: int = 0
    model_version: str = "n/a"
    metadata: Dict[str, str] = field(default_factory=lambda: {})
    debug: bool = False

    def __repr__(self):
        return (
            f"name: {self.name}, "
            f"events: {[s.__repr__() for s in self.events]}, "
            f"input_sample_rate: {self.input_sample_rate}, "
            f"samples_per_window: {self.samples_per_window}, "
            f"samples_per_hop: {self.samples_per_hop}, "
            f"model_version: {self.model_version}"
        )

    def __str__(self):
        return (
            f"name: {self.name}, "
            f"events: {[s.__str__() for s in self.events]}, "
            f"input_sample_rate: {self.input_sample_rate}, "
            f"samples_per_window: {self.samples_per_window}, "
            f"samples_per_hop: {self.samples_per_hop}, "
            f"model_version: {self.model_version}"
        )

    def as_dict(self) -> dict:
        """
        :return: EventStream as a dictionary
        """
        return {
            "name": self.name,
            "events": [e.as_dict() for e in self.events],
            "input_sample_rate": self.input_sample_rate,
            "samples_per_window": self.samples_per_window,
            "samples_per_hop": self.samples_per_hop,
            "model_version": self.model_version,
            "metadata": self.metadata,
        }

    def has_data(self):
        """
        :return: if there is at least one event
        """
        return len(self.events) > 0

    def has_events(self) -> bool:
        """
        :return: True if there are one or more events in the stream
        """
        return len(self.events) > 0

    def get_event(self, index: int = 0) -> Optional[Event]:
        """
        :param index: index of event to get.  Use negative values to select from the end of the list.
                        Default 0 (first event)
        :return: Event at the index, or None if the event/index doesn't exist
        """
        if 0 > index:
            index += len(self.events)
        if 0 <= index < len(self.events):
            return self.events[index]
        return None

    def get_data_column(self, column_name: str) -> list:
        """
        return a list of data with key column_name from each of the events
        if column_name doesn't exist, gets a list of valid column_names

        :param column_name: key of data to get
        :return: list of data named column_name or the list of all possible column names
        """
        result = []
        column_list = set()
        for r in self.events:
            val = r.get_item(column_name)
            if type(val) != list:
                result.append(val)
            else:
                for v in val:
                    column_list.add(v)
        if len(result) > 0:
            return result
        return list(column_list)

    @staticmethod
    def from_eventstream(
        stream: RedvoxPacketM.EventStream, save_mode: FileSystemSaveMode = FileSystemSaveMode.MEM, base_dir: str = "."
    ) -> "EventStream":
        """
        convert a Redvox Api1000 Packet EventStream into its sdk version

        :param stream: Redvox Api1000 Packet EventStream to read data from
        :param save_mode: FileSystemSaveMode that determines how Event data is saved.
                            Default FileSystemSaveMode.MEM (use RAM).  Other options are DISK (save to directory)
                            and TEMP (save to temporary directory)
        :param base_dir: the location of the parquet file that holds the Event data.  Not used if save_data is False.
                            Default current directory (".")
        :return: EventStream (sdk version)
        """
        result = EventStream(stream.name, metadata=dict(stream.metadata))
        if "input_sample_rate" in stream.metadata.keys():
            result.input_sample_rate = int(stream.metadata.get("input_sample_rate"))
        if "input_samples_per_window" in stream.metadata.keys():
            result.samples_per_window = int(stream.metadata.get("input_samples_per_window"))
        if "input_samples_per_hop" in stream.metadata.keys():
            result.samples_per_hop = int(stream.metadata.get("input_samples_per_hop"))
        if "model_version" in stream.metadata.keys():
            result.model_version = stream.metadata.get("model_version")
        result.add_events(stream, save_mode=save_mode, base_dir=base_dir)
        return result

    def add_events(
        self,
        stream: RedvoxPacketM.EventStream,
        save_mode: FileSystemSaveMode = FileSystemSaveMode.MEM,
        base_dir: str = ".",
    ):
        """
        add events from a Redvox Api1000 Packet EventStream with the same name.
        Does nothing if names do not match

        :param stream: stream of events to add
        :param save_mode: FileSystemSaveMode that determines how Event data is saved.
                            Default FileSystemSaveMode.MEM (use RAM).  Other options are DISK (save to directory)
                            and TEMP (save to temporary directory)
        :param base_dir: the location of the parquet file that holds the Event data.  Not used if save_data is False.
                            Default current directory (".")
        """
        if self.name == stream.name:
            timestamps = stream.timestamps.timestamps
            events = stream.events
            for i in range(len(timestamps)):
                self.events.append(Event(timestamps[i], save_mode=save_mode, base_dir=base_dir).read_raw(events[i]))
        elif self.debug:
            print(f"Stream name mismatch while adding to EventStream.  Expected {self.name}, got {stream.name}.")

    def sort_events(self, asc: bool = True):
        """
        sort the events in the stream via ascending or descending timestamp order

        :param asc: if True, data is sorted in ascending order
        """
        self.events.sort(key=lambda e: e.get_timestamp(), reverse=not asc)

    def num_events(self) -> int:
        """
        :return: number of events in stream
        """
        return len(self.events)

    def sample_rate_hz(self):
        """
        :return: sample rate of events in the stream in hz
        """
        return np.mean(np.diff([e.get_timestamp() for e in self.events]))

    def window_sample_rate_hz(self):
        """
        :return: idealized event sample window rate in hz
        """
        return self.input_sample_rate / self.samples_per_window

    def hop_sample_rate_hz(self):
        """
        :return: idealized event sample hop rate in hz
        """
        return self.input_sample_rate / self.samples_per_hop

    def create_event_window(self, start: float = -np.inf, end: float = np.inf):
        """
        removes any event in the stream that doesn't match start <= event < end
        adds empty events to beginning and end of data (as long as the corresponding input values are not infinity)
        default start is negative infinity, default end is infinity
        all times in microseconds since epoch UTC

        :param start: inclusive start time of events to keep
        :param end: exclusive end time of events to keep
        """
        self.events = [s for s in self.events if start <= s.get_timestamp() < end]
        if self.num_events() > 0:
            if start < self.events[0].get_timestamp() and not np.isinf(start):
                self.events.insert(0, Event(start, self.name))
            if not np.isinf(end):
                self.events.append(Event(end - 1, self.name))

    def get_file_names(self) -> List[str]:
        """
        :return: the names of the files which store the event data
        """
        return [e.file_name() for e in self.events]

    def save_streams(self):
        """
        saves all streams to disk

        note: use the function set_save_dir() to change where events are saved
        """
        for e in self.events:
            if e.is_save_to_disk():
                e.to_json_file()

    def set_save_dir(self, new_dir: str):
        """
        change the directory where events are saved to

        :param new_dir: new directory path
        """
        for e in self.events:
            e.set_save_dir(new_dir)

    def set_save_mode(self, new_save_mode: FileSystemSaveMode):
        """
        update the save mode for all EventStream

        :param new_save_mode: save mode to set
        """
        for e in self.events:
            e.set_save_mode(new_save_mode)

    def update_timestamps(self, offset_model: om.OffsetModel, use_model_function: bool = False):
        """
        update the timestamps in the data

        :param offset_model: model used to update the timestamps
        :param use_model_function: if True, use the model's slope function to update the timestamps.
                                    otherwise uses the best offset (model's intercept value).  Default False
        """
        for evnt in self.events:
            evnt.update_timestamps(offset_model, use_model_function)

    def original_timestamps(self, offset_model: om.OffsetModel, use_model_function: bool = False):
        """
        undo the update to the timestamps in the data

        :param offset_model: model used to update the timestamps
        :param use_model_function: if True, use the model's slope function to update the timestamps.
                                    otherwise uses the best offset (model's intercept value).  Default False
        """
        for evnt in self.events:
            evnt.original_timestamps(offset_model, use_model_function)

    @staticmethod
    def from_json_dict(json_dict: dict) -> "EventStream":
        """
        :param json_dict: json dict to parse
        :return: EventStream from json dict
        """
        if "name" in json_dict.keys():
            result = EventStream(
                json_dict["name"],
                [Event.from_json_dict(e) for e in json_dict["events"]],
                json_dict["input_sample_rate"],
                json_dict["samples_per_window"],
                json_dict["samples_per_hop"],
                json_dict["model_version"],
                json_dict["metadata"],
            )
        else:
            result = EventStream("Empty Stream; no name for identification")
        return result

    @staticmethod
    def from_json_file(file_dir: str, file_name: str) -> "EventStream":
        """
        :param file_dir: full path to containing directory for the file
        :param file_name: name of file to load data from
        :return: EventStream from json file
        """
        json_data = json_file_to_dict(os.path.join(file_dir, f"{file_name}"))
        if "name" in json_data.keys():
            result = EventStream(
                json_data["name"],
                json_data["events"],
                json_data["input_sample_rate"],
                json_data["samples_per_window"],
                json_data["samples_per_hop"],
                json_data["model_version"],
                json_data["metadata"],
            )
            result.set_save_mode(FileSystemSaveMode.DISK)
            result.set_save_dir(file_dir)
        else:
            result = EventStream("Empty Stream; no name for identification")
        return result

    def to_json_file(self, file_dir: str = ".", file_name: Optional[str] = None) -> Path:
        """
        saves the EventStream as a json file

        :param file_dir: the directory to save the file into.  default current directory (".")
        :param file_name: the optional base file name.  Do not include a file extension.
                            If None, a default file name is created using this format:
                            eventstream_[eventstream.name].json
        :return: path to json file
        """
        return io.eventstream_to_json_file(self, file_dir, file_name)

    def print_errors(self):
        """
        print all errors to screen
        """
        for e in self.events:
            e.print_errors()


@dataclass_json
@dataclass
class EventStreams:
    """
    stores multiple event streams per station.

    ALL timestamps in microseconds since epoch UTC unless otherwise stated

    Properties:
        streams: List[EventStream]; list of all EventStream.  Default empty list

        ml_data: Optional ExtractedMl from the packets.  Default None

        debug: bool; if True, output additional information during runtime.  Default False
    """

    streams: List[EventStream] = field(default_factory=lambda: [])
    ml_data: Optional[ml.ExtractedMl] = None
    debug: bool = False

    def __repr__(self):
        return f"streams: {[s.__repr__() for s in self.streams]}, ml_data: {self.ml_data}, debug: {self.debug}"

    def __str__(self):
        return f"streams: {[s.__str__() for s in self.streams]}, ml_data: {self.ml_data}"

    def as_dict(self) -> dict:
        """
        :return: EventStreams as dict
        """
        return {
            "streams": [s.as_dict() for s in self.streams],
            "ml_data": self.ml_data.to_dict() if self.ml_data else None,
        }

    def read_from_packet(self, packet: RedvoxPacketM):
        """
        read the eventstream payload from a single Redvox Api1000 packet

        :param packet: packet to read data from
        """
        for st in packet.event_streams:
            if st.name == ml.ML_EVENT_STREAM_NAME:
                if self.ml_data:
                    self.ml_data.windows.extend(ml.extract_ml_windows(_find_ml_event_stream(packet)))
                else:
                    self.ml_data = ml.extract_ml_from_packet(WrappedRedvoxPacketM(packet))
            else:
                if st.name in self.get_stream_names() and self.get_stream(st.name).has_data():
                    self.get_stream(st.name).add_events(st)
                else:
                    self.remove_stream(st.name)
                    self.streams.append(EventStream.from_eventstream(st))

    def read_from_packets_list(self, packets: List[RedvoxPacketM]):
        """
        read the eventstream payload from multiple Redvox Api1000 packets

        :param packets: packets to read data from
        """
        for p in packets:
            if type(p) == RedvoxPacketM:
                self.read_from_packet(p)

    def append(self, other_stream: EventStream):
        """
        append another EventStream to an existing EventStream or add to the list of EventStream

        :param other_stream: other EventStream to add
        """
        if other_stream.name in self.get_stream_names():
            self.get_stream(other_stream.name).add_events(other_stream)
        else:
            self.streams.append(other_stream)

    def append_ml(self, other_ml: ml.ExtractedMl):
        """
        append the windows from another extracted machine learning object or
        set the existing ML object if its empty.

        :param other_ml: other ExtractedMl to add
        """
        if self.ml_data:
            self.ml_data.windows.extend(other_ml.windows)
        else:
            self.ml_data = other_ml

    def append_streams(self, other_streams: "EventStreams"):
        """
        append another EventStreams object to an existing EventStreams object

        :param other_streams: EventStreams to add
        """
        for s in other_streams.streams:
            self.append(s)
        self.append_ml(other_streams.ml_data)

    def remove_stream(self, stream_name: str):
        """
        remove any stream with the same stream_name

        :param stream_name: name of stream to remove
        """
        self.streams = [s for s in self.streams if s.name != stream_name]

    def get_stream(self, stream_name: str) -> Optional[EventStream]:
        """
        :param stream_name: name of event stream to get
        :return: the EventStream that has the name specified or None if it doesn't exist
        """
        for s in self.streams:
            if s.name == stream_name:
                return s
        if self.debug:
            print(f"{stream_name} does not exist in streams.  Use one of {[self.get_stream_names()]}")
        return None

    def get_stream_names(self) -> List[str]:
        """
        :return: names of all streams
        """
        return [s.name for s in self.streams]

    def create_event_window(self, start: float = -np.inf, end: float = np.inf):
        """
        removes any event in the streams and ML that doesn't match start <= event < end
        default start is negative infinity, default end is infinity
        all times in microseconds since epoch UTC

        :param start: inclusive start time of events to keep
        :param end: exclusive end time of events to keep
        """
        for s in self.streams:
            s.create_event_window(start, end)
        if self.ml_data:
            self.ml_data.windows = [s for s in self.ml_data.windows if start <= s.timestamp < end]

    def set_save_dir(self, new_dir: str):
        """
        change the directory where events are saved to

        :param new_dir: new directory path
        """
        for s in self.streams:
            s.set_save_dir(new_dir)

    def set_save_mode(self, new_save_mode: FileSystemSaveMode):
        """
        update the save mode for all EventStream

        :param new_save_mode: save mode to set
        """
        for s in self.streams:
            s.set_save_mode(new_save_mode)

    def update_timestamps(self, offset_model: om.OffsetModel, use_model_function: bool = False):
        """
        update the timestamps in the data

        :param offset_model: model used to update the timestamps
        :param use_model_function: if True, use the model's slope function to update the timestamps.
                                    otherwise uses the best offset (model's intercept value).  Default False
        """
        for evnt in self.streams:
            evnt.update_timestamps(offset_model, use_model_function)
        if self.ml_data:
            for w in self.ml_data.windows:
                w.timestamp = offset_model.update_time(w.timestamp, use_model_function)

    def original_timestamps(self, offset_model: om.OffsetModel, use_model_function: bool = False):
        """
        undo the update to the timestamps in the data

        :param offset_model: model used to update the timestamps
        :param use_model_function: if True, use the model's slope function to update the timestamps.
                                    otherwise uses the best offset (model's intercept value).  Default False
        """
        for evnt in self.streams:
            evnt.original_timestamps(offset_model, use_model_function)
        if self.ml_data:
            for w in self.ml_data.windows:
                w.timestamp = offset_model.get_original_time(w.timestamp, use_model_function)

    @staticmethod
    def from_dict(in_dict: dict) -> "EventStreams":
        """
        :param in_dict: dictionary representing an EventStreams object
        :return: the EventStreams object from the dictionary
        """
        result = EventStreams()
        if "streams" in in_dict.keys():
            result.streams = [EventStream.from_json_dict(s) for s in in_dict["streams"]]
        if "ml_data" in in_dict.keys():
            result.ml_data = ml.ExtractedMl.from_dict(in_dict["ml_data"])
        return result

    @staticmethod
    def from_json_file(file_dir: str, file_name: str) -> "EventStreams":
        """
        :param file_dir: full path to containing directory for the file
        :param file_name: name of file to load data from
        :return: EventStreams from json file
        """
        return EventStreams.from_dict(json_file_to_dict(os.path.join(file_dir, f"{file_name}")))

    def to_json_file(self, file_dir: str = ".", file_name: Optional[str] = None) -> Path:
        """
        saves the EventStream as a json file

        :param file_dir: the directory to save the file into.  default current directory (".")
        :param file_name: the optional base file name.  Do not include a file extension.
                            If None, a default file name is created using this format:
                            eventstreams.json
        :return: path to json file
        """
        return io.eventstreams_to_json_file(self, file_dir, file_name)


def _find_ml_event_stream(packet: RedvoxPacketM) -> Optional[es.EventStream]:
    """
    Attempts to find an event stream with ML data.

    :param packet: The packet to search in.
    :return: An instance of the matching event stream or None.
    """
    stream: RedvoxPacketM.EventStream
    for stream in packet.event_streams:
        if stream.name == ml.ML_EVENT_STREAM_NAME:
            return es.EventStream(stream)
    return None


def _get_ml_from_packet(packet: RedvoxPacketM) -> Optional[ml.ExtractedMl]:
    """
    reads the machine learning payload from a single Redvox Api1000 packet

    :param packet: packet to read machine learning data from
    """
    stream: Optional[es.EventStream] = _find_ml_event_stream(packet)
    return None if stream is None else ml.extract_ml_from_event_stream(stream)



"""
This module provides IO primitives for working with data windows.
"""
from dataclasses import dataclass
import os.path
from pathlib import Path
import pickle
import json
import enum

from typing import (
    Dict,
    List,
    Optional,
    TYPE_CHECKING,
)

import lz4.frame

from redvox.common.io import FileSystemWriter, FileSystemSaveMode, json_to_dict


if TYPE_CHECKING:
    from redvox.common.data_window import DataWindow


class DataWindowOutputType(enum.Enum):
    """
    Type of file to create when exporting DataWindow
    """

    NONE: int = 0
    LZ4: int = 1
    PARQUET: int = 2
    JSON: int = 3

    @staticmethod
    def list_names() -> List[str]:
        """
        :return: list of possible values for OutputType
        """
        return [n.name for n in DataWindowOutputType]

    @staticmethod
    def list_non_none_names() -> List[str]:
        """
        :return: List of possible non-None values for OutputType
        """
        return [n.name for n in DataWindowOutputType if n != DataWindowOutputType.NONE]

    @staticmethod
    def str_to_type(str_type: str) -> "DataWindowOutputType":
        """
        converts the string to the corresponding OutputType
        if the type given is not in list_non_none_names(), returns NONE value

        :param str_type: string to convert
        :return: DataWindowOutputType matching string given or NONE
        """
        str_type = str_type.upper()
        if str_type in DataWindowOutputType.list_non_none_names():
            return DataWindowOutputType[str_type]
        return DataWindowOutputType["NONE"]


class DataWindowFileSystemWriter(FileSystemWriter):
    """
    This class holds the FileSystemWriter info for DataWindows.  Extends the FileSystemWriter from io.py

    Properties:
        file_name: str, the name of the file (do not include extension)

        file_ext: str, the extension used by the file (do not include the .)  Default "NONE"

        base_dir: str, the directory to save the file to.  Default "." (current dir)

        make_run_me: bool, if True, makes a sample runme.py file when saving to disk.  default False

        orig_path: str, the current working directory when the object is initialized

    Protected:
        _save_mode: FileSystemSaveMode, determines how files get saved

        _temp_dir: TemporaryDirectory, temporary directory for large files when not saving to disk
    """

    def __init__(self, file_name: str, file_ext: str = "none", base_dir: str = ".", make_run_me: bool = False):
        """
        initialize DataWindowFileSystemWriter

        :param file_name: name of file
        :param file_ext: extension of file, default "none"
        :param base_dir: directory to save file to, default "." (current dir)
        :param make_run_me: if True, add a runme.py file to the saved files.  Default False
        """
        self.orig_path = os.getcwd()
        if not os.path.exists(base_dir):
            os.makedirs(base_dir, exist_ok=True)
        os.chdir(base_dir)
        super().__init__(
            file_name,
            file_ext,
            ".",
            FileSystemSaveMode.DISK
            if DataWindowOutputType.str_to_type(file_ext) != DataWindowOutputType.NONE
            else FileSystemSaveMode.MEM,
        )
        self.make_run_me = make_run_me

    def set_extension(self, ext: str):
        """
        change the file extension.  Valid values are "PARQUET", "LZ4", "JSON" and "NONE".  Invalid values become "NONE"

        :param ext: extension to change to
        """
        self.file_extension = DataWindowOutputType.str_to_type(ext).name.lower()


@dataclass
class DataWindowSerializationResult:
    path: str
    serialized_bytes: int
    compressed_bytes: int


def data_window_as_json(data_window: "DataWindow") -> str:
    """
    Converts the DataWindow's metadata into a JSON dictionary

    :param data_window: The data window to convert
    :return: The data window's metadata as a JSON dictionary
    """
    return json.dumps(data_window.as_dict())


def data_window_to_json(
    data_window: "DataWindow",
    base_dir: str = ".",
    file_name: Optional[str] = None,
) -> Path:
    """
    Converts the DataWindow into a JSON metadata file

    :param data_window: The data window to convert.
    :param base_dir: The base directory to write the JSON file to (default=.).
    :param file_name: The optional file name. If None, a default filename with the following format is used:
                      [event_name].json
    :return: The path to the written metadata file.
    """
    _file_name = file_name if file_name is not None else data_window.event_name
    os.makedirs(base_dir, exist_ok=True)
    for s in data_window.stations():
        s.to_json_file()
    file_path: Path = Path(base_dir).joinpath(f"{_file_name}.json")
    with open(file_path, "w") as f:
        f.write(data_window_as_json(data_window))
        return file_path.resolve(False)


def json_file_to_data_window(file_path: str) -> Dict:
    """
    load a specifically named DataWindow as a dictionary from a directory

    :param file_path: full path of file to load
    :return: the dictionary of the DataWindow if it exists, or None otherwise
    """
    with open(file_path, "r") as f_p:
        return json_to_dict(f_p.read())


def serialize_data_window(
    data_window: "DataWindow",
    base_dir: str = ".",
    file_name: Optional[str] = None,
    compression_factor: int = 4,
) -> Path:
    """
    Serializes and compresses a DataWindow to a file and creates a JSON metadata file for the compressed file.

    :param data_window: The data window to serialize and compress.
    :param base_dir: The base directory to write the serialized file to (default=.).
    :param file_name: The optional file name. If None, a default filename with the following format is used:
                      [start_ts]_[end_ts]_[event_name].pkl.lz4
    :param compression_factor: A value between 1 and 12. Higher values provide better compression, but take longer.
                               (default=4).
    :return: The path to the written compressed file.
    """

    _file_name: str = (
        file_name
        if file_name is not None
        else f"{int(data_window.start_date())}"
        f"_{int(data_window.end_date())}"
        f"_{len(data_window.event_name)}.pkl.lz4"
    )

    json_path: Path = data_window.fs_writer().json_path()
    with open(json_path, "w") as f:
        f.write(data_window.to_json())
        json_path.resolve(False)

    file_path: Path = Path(base_dir).joinpath(_file_name)

    with lz4.frame.open(file_path, "wb", compression_level=compression_factor) as compressed_out:
        pickle.dump(data_window, compressed_out)
        compressed_out.flush()
        return file_path.resolve(False)


def deserialize_data_window(path: str) -> "DataWindow":
    """
    Decompresses and deserializes a DataWindow written to disk.

    :param path: Path to the serialized and compressed data window.
    :return: An instance of a DataWindow.
    """
    with lz4.frame.open(path, "rb") as compressed_in:
        return pickle.load(compressed_in)



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

ACCELEROMETER_X: ChannelType
ACCELEROMETER_Y: ChannelType
ACCELEROMETER_Z: ChannelType
ACCURACY: ChannelType
ALTITUDE: ChannelType
BAROMETER: ChannelType
DESCRIPTOR: _descriptor.FileDescriptor
GYROSCOPE_X: ChannelType
GYROSCOPE_Y: ChannelType
GYROSCOPE_Z: ChannelType
IMAGE: ChannelType
INFRARED: ChannelType
LATITUDE: ChannelType
LIGHT: ChannelType
LONGITUDE: ChannelType
MAGNETOMETER_X: ChannelType
MAGNETOMETER_Y: ChannelType
MAGNETOMETER_Z: ChannelType
MICROPHONE: ChannelType
OTHER: ChannelType
RESERVED_0: ChannelType
RESERVED_1: ChannelType
RESERVED_2: ChannelType
SPEED: ChannelType
TIME_SYNCHRONIZATION: ChannelType

class BytePayload(_message.Message):
    __slots__ = ["bytePayloadType", "payload"]
    class BytePayloadType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    BYTEPAYLOADTYPE_FIELD_NUMBER: _ClassVar[int]
    BYTES: BytePayload.BytePayloadType
    FLOAT32: BytePayload.BytePayloadType
    FLOAT64: BytePayload.BytePayloadType
    INT16: BytePayload.BytePayloadType
    INT24: BytePayload.BytePayloadType
    INT32: BytePayload.BytePayloadType
    INT64: BytePayload.BytePayloadType
    INT8: BytePayload.BytePayloadType
    OTHER: BytePayload.BytePayloadType
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    UINT32: BytePayload.BytePayloadType
    UINT64: BytePayload.BytePayloadType
    UINT8: BytePayload.BytePayloadType
    UNINT16: BytePayload.BytePayloadType
    UNINT24: BytePayload.BytePayloadType
    bytePayloadType: BytePayload.BytePayloadType
    payload: bytes
    def __init__(self, bytePayloadType: _Optional[_Union[BytePayload.BytePayloadType, str]] = ..., payload: _Optional[bytes] = ...) -> None: ...

class EvenlySampledChannel(_message.Message):
    __slots__ = ["byte_payload", "channel_types", "first_sample_timestamp_epoch_microseconds_utc", "float32_payload", "float64_payload", "int32_payload", "int64_payload", "metadata", "sample_rate_hz", "sensor_name", "uint32_payload", "uint64_payload", "value_means", "value_medians", "value_stds"]
    BYTE_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    CHANNEL_TYPES_FIELD_NUMBER: _ClassVar[int]
    FIRST_SAMPLE_TIMESTAMP_EPOCH_MICROSECONDS_UTC_FIELD_NUMBER: _ClassVar[int]
    FLOAT32_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    FLOAT64_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    INT32_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    INT64_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    SAMPLE_RATE_HZ_FIELD_NUMBER: _ClassVar[int]
    SENSOR_NAME_FIELD_NUMBER: _ClassVar[int]
    UINT32_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    UINT64_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    VALUE_MEANS_FIELD_NUMBER: _ClassVar[int]
    VALUE_MEDIANS_FIELD_NUMBER: _ClassVar[int]
    VALUE_STDS_FIELD_NUMBER: _ClassVar[int]
    byte_payload: BytePayload
    channel_types: _containers.RepeatedScalarFieldContainer[ChannelType]
    first_sample_timestamp_epoch_microseconds_utc: int
    float32_payload: Float32Payload
    float64_payload: Float64Payload
    int32_payload: Int32Payload
    int64_payload: Int64Payload
    metadata: _containers.RepeatedScalarFieldContainer[str]
    sample_rate_hz: float
    sensor_name: str
    uint32_payload: UInt32Payload
    uint64_payload: UInt64Payload
    value_means: _containers.RepeatedScalarFieldContainer[float]
    value_medians: _containers.RepeatedScalarFieldContainer[float]
    value_stds: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, channel_types: _Optional[_Iterable[_Union[ChannelType, str]]] = ..., sensor_name: _Optional[str] = ..., sample_rate_hz: _Optional[float] = ..., first_sample_timestamp_epoch_microseconds_utc: _Optional[int] = ..., byte_payload: _Optional[_Union[BytePayload, _Mapping]] = ..., uint32_payload: _Optional[_Union[UInt32Payload, _Mapping]] = ..., uint64_payload: _Optional[_Union[UInt64Payload, _Mapping]] = ..., int32_payload: _Optional[_Union[Int32Payload, _Mapping]] = ..., int64_payload: _Optional[_Union[Int64Payload, _Mapping]] = ..., float32_payload: _Optional[_Union[Float32Payload, _Mapping]] = ..., float64_payload: _Optional[_Union[Float64Payload, _Mapping]] = ..., value_means: _Optional[_Iterable[float]] = ..., value_stds: _Optional[_Iterable[float]] = ..., value_medians: _Optional[_Iterable[float]] = ..., metadata: _Optional[_Iterable[str]] = ...) -> None: ...

class Float32Payload(_message.Message):
    __slots__ = ["payload"]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    payload: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, payload: _Optional[_Iterable[float]] = ...) -> None: ...

class Float64Payload(_message.Message):
    __slots__ = ["payload"]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    payload: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, payload: _Optional[_Iterable[float]] = ...) -> None: ...

class Int32Payload(_message.Message):
    __slots__ = ["payload"]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    payload: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, payload: _Optional[_Iterable[int]] = ...) -> None: ...

class Int64Payload(_message.Message):
    __slots__ = ["payload"]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    payload: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, payload: _Optional[_Iterable[int]] = ...) -> None: ...

class RedvoxPacket(_message.Message):
    __slots__ = ["acquisition_server", "api", "app_file_start_timestamp_epoch_microseconds_utc", "app_file_start_timestamp_machine", "app_version", "authenticated_email", "authentication_server", "authentication_token", "battery_level_percent", "device_make", "device_model", "device_os", "device_os_version", "device_temperature_c", "evenly_sampled_channels", "firebase_token", "is_backfilled", "is_private", "is_scrambled", "metadata", "redvox_id", "server_timestamp_epoch_microseconds_utc", "time_synchronization_server", "unevenly_sampled_channels", "uuid"]
    ACQUISITION_SERVER_FIELD_NUMBER: _ClassVar[int]
    API_FIELD_NUMBER: _ClassVar[int]
    APP_FILE_START_TIMESTAMP_EPOCH_MICROSECONDS_UTC_FIELD_NUMBER: _ClassVar[int]
    APP_FILE_START_TIMESTAMP_MACHINE_FIELD_NUMBER: _ClassVar[int]
    APP_VERSION_FIELD_NUMBER: _ClassVar[int]
    AUTHENTICATED_EMAIL_FIELD_NUMBER: _ClassVar[int]
    AUTHENTICATION_SERVER_FIELD_NUMBER: _ClassVar[int]
    AUTHENTICATION_TOKEN_FIELD_NUMBER: _ClassVar[int]
    BATTERY_LEVEL_PERCENT_FIELD_NUMBER: _ClassVar[int]
    DEVICE_MAKE_FIELD_NUMBER: _ClassVar[int]
    DEVICE_MODEL_FIELD_NUMBER: _ClassVar[int]
    DEVICE_OS_FIELD_NUMBER: _ClassVar[int]
    DEVICE_OS_VERSION_FIELD_NUMBER: _ClassVar[int]
    DEVICE_TEMPERATURE_C_FIELD_NUMBER: _ClassVar[int]
    EVENLY_SAMPLED_CHANNELS_FIELD_NUMBER: _ClassVar[int]
    FIREBASE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    IS_BACKFILLED_FIELD_NUMBER: _ClassVar[int]
    IS_PRIVATE_FIELD_NUMBER: _ClassVar[int]
    IS_SCRAMBLED_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    REDVOX_ID_FIELD_NUMBER: _ClassVar[int]
    SERVER_TIMESTAMP_EPOCH_MICROSECONDS_UTC_FIELD_NUMBER: _ClassVar[int]
    TIME_SYNCHRONIZATION_SERVER_FIELD_NUMBER: _ClassVar[int]
    UNEVENLY_SAMPLED_CHANNELS_FIELD_NUMBER: _ClassVar[int]
    UUID_FIELD_NUMBER: _ClassVar[int]
    acquisition_server: str
    api: int
    app_file_start_timestamp_epoch_microseconds_utc: int
    app_file_start_timestamp_machine: int
    app_version: str
    authenticated_email: str
    authentication_server: str
    authentication_token: str
    battery_level_percent: float
    device_make: str
    device_model: str
    device_os: str
    device_os_version: str
    device_temperature_c: float
    evenly_sampled_channels: _containers.RepeatedCompositeFieldContainer[EvenlySampledChannel]
    firebase_token: str
    is_backfilled: bool
    is_private: bool
    is_scrambled: bool
    metadata: _containers.RepeatedScalarFieldContainer[str]
    redvox_id: str
    server_timestamp_epoch_microseconds_utc: int
    time_synchronization_server: str
    unevenly_sampled_channels: _containers.RepeatedCompositeFieldContainer[UnevenlySampledChannel]
    uuid: str
    def __init__(self, api: _Optional[int] = ..., uuid: _Optional[str] = ..., redvox_id: _Optional[str] = ..., authenticated_email: _Optional[str] = ..., authentication_token: _Optional[str] = ..., firebase_token: _Optional[str] = ..., is_backfilled: bool = ..., is_private: bool = ..., is_scrambled: bool = ..., device_make: _Optional[str] = ..., device_model: _Optional[str] = ..., device_os: _Optional[str] = ..., device_os_version: _Optional[str] = ..., app_version: _Optional[str] = ..., battery_level_percent: _Optional[float] = ..., device_temperature_c: _Optional[float] = ..., acquisition_server: _Optional[str] = ..., time_synchronization_server: _Optional[str] = ..., authentication_server: _Optional[str] = ..., app_file_start_timestamp_epoch_microseconds_utc: _Optional[int] = ..., app_file_start_timestamp_machine: _Optional[int] = ..., server_timestamp_epoch_microseconds_utc: _Optional[int] = ..., evenly_sampled_channels: _Optional[_Iterable[_Union[EvenlySampledChannel, _Mapping]]] = ..., unevenly_sampled_channels: _Optional[_Iterable[_Union[UnevenlySampledChannel, _Mapping]]] = ..., metadata: _Optional[_Iterable[str]] = ...) -> None: ...

class RedvoxPacketResponse(_message.Message):
    __slots__ = ["checksum", "errors", "metadata", "type"]
    class Error(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    class Type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    CHECKSUM_FIELD_NUMBER: _ClassVar[int]
    ERROR: RedvoxPacketResponse.Type
    ERRORS_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    NOT_AUTHENTICATED: RedvoxPacketResponse.Error
    OK: RedvoxPacketResponse.Type
    OTHER: RedvoxPacketResponse.Error
    TYPE_FIELD_NUMBER: _ClassVar[int]
    checksum: int
    errors: _containers.RepeatedScalarFieldContainer[RedvoxPacketResponse.Error]
    metadata: _containers.RepeatedScalarFieldContainer[str]
    type: RedvoxPacketResponse.Type
    def __init__(self, type: _Optional[_Union[RedvoxPacketResponse.Type, str]] = ..., checksum: _Optional[int] = ..., errors: _Optional[_Iterable[_Union[RedvoxPacketResponse.Error, str]]] = ..., metadata: _Optional[_Iterable[str]] = ...) -> None: ...

class UInt32Payload(_message.Message):
    __slots__ = ["payload"]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    payload: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, payload: _Optional[_Iterable[int]] = ...) -> None: ...

class UInt64Payload(_message.Message):
    __slots__ = ["payload"]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    payload: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, payload: _Optional[_Iterable[int]] = ...) -> None: ...

class UnevenlySampledChannel(_message.Message):
    __slots__ = ["byte_payload", "channel_types", "float32_payload", "float64_payload", "int32_payload", "int64_payload", "metadata", "sample_interval_mean", "sample_interval_median", "sample_interval_std", "sensor_name", "timestamps_microseconds_utc", "uint32_payload", "uint64_payload", "value_means", "value_medians", "value_stds"]
    BYTE_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    CHANNEL_TYPES_FIELD_NUMBER: _ClassVar[int]
    FLOAT32_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    FLOAT64_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    INT32_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    INT64_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    SAMPLE_INTERVAL_MEAN_FIELD_NUMBER: _ClassVar[int]
    SAMPLE_INTERVAL_MEDIAN_FIELD_NUMBER: _ClassVar[int]
    SAMPLE_INTERVAL_STD_FIELD_NUMBER: _ClassVar[int]
    SENSOR_NAME_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMPS_MICROSECONDS_UTC_FIELD_NUMBER: _ClassVar[int]
    UINT32_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    UINT64_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    VALUE_MEANS_FIELD_NUMBER: _ClassVar[int]
    VALUE_MEDIANS_FIELD_NUMBER: _ClassVar[int]
    VALUE_STDS_FIELD_NUMBER: _ClassVar[int]
    byte_payload: BytePayload
    channel_types: _containers.RepeatedScalarFieldContainer[ChannelType]
    float32_payload: Float32Payload
    float64_payload: Float64Payload
    int32_payload: Int32Payload
    int64_payload: Int64Payload
    metadata: _containers.RepeatedScalarFieldContainer[str]
    sample_interval_mean: float
    sample_interval_median: float
    sample_interval_std: float
    sensor_name: str
    timestamps_microseconds_utc: _containers.RepeatedScalarFieldContainer[int]
    uint32_payload: UInt32Payload
    uint64_payload: UInt64Payload
    value_means: _containers.RepeatedScalarFieldContainer[float]
    value_medians: _containers.RepeatedScalarFieldContainer[float]
    value_stds: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, channel_types: _Optional[_Iterable[_Union[ChannelType, str]]] = ..., sensor_name: _Optional[str] = ..., timestamps_microseconds_utc: _Optional[_Iterable[int]] = ..., byte_payload: _Optional[_Union[BytePayload, _Mapping]] = ..., uint32_payload: _Optional[_Union[UInt32Payload, _Mapping]] = ..., uint64_payload: _Optional[_Union[UInt64Payload, _Mapping]] = ..., int32_payload: _Optional[_Union[Int32Payload, _Mapping]] = ..., int64_payload: _Optional[_Union[Int64Payload, _Mapping]] = ..., float32_payload: _Optional[_Union[Float32Payload, _Mapping]] = ..., float64_payload: _Optional[_Union[Float64Payload, _Mapping]] = ..., sample_interval_mean: _Optional[float] = ..., sample_interval_std: _Optional[float] = ..., sample_interval_median: _Optional[float] = ..., value_means: _Optional[_Iterable[float]] = ..., value_stds: _Optional[_Iterable[float]] = ..., value_medians: _Optional[_Iterable[float]] = ..., metadata: _Optional[_Iterable[str]] = ...) -> None: ...

class ChannelType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
